#  Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



class StimSeq :
    stimSeq     = None # [ nEvent x nSymb ] stimulus code for each time point for each stimulus
    stimTime_ms = None # time stim i ends, i.e. stimulus i is on screen from stimTime_ms[i-1]-stimTime_ms[i]
    eventSeq    = None # events to send at each stimulus point

    def __init__(self,st=None,ss=None,es=None):
        self.stimSeq     = ss
        self.stimTime_ms = st
        self.eventSeq    = es

    def __str__(self):
        res = "#stimTimes: ";
        if not self.stimTime_ms is None:
            res += "(1," + str(len(self.stimTime_ms)) + ")\n"
            for i in range(0,len(self.stimTime_ms)-1):
                res += str(self.stimTime_ms[i]) + " "
            res+= str(self.stimTime_ms[-1]) + "\n"
        else:
            res+= "<null>\n"
        res+= "\n\n"
        res += "#stimSeq : "
        if not self.stimSeq is None:
            res += "(" + str(len(self.stimSeq[0])) + "," + str(len(self.stimSeq)) + ")\n"
            for j in range(0,len(self.stimSeq[0])):
                for i in range(len(self.stimSeq)-1):
                    res += str(self.stimSeq[i][j]) + " "                
                res+= str(self.stimSeq[-1][j]) + "\n"
        else:
            res+= "<null>\n"
        res+="\n\n"
        return res

    def convertstimSeq2int(self,scale=1):
        self.stimSeq = self.float2int(self.stimSeq)

    @staticmethod
    def float2int(stimSeq,scale=1,minval=None,maxval=None):
        '''convert float list of lists to integer'''
        if type(stimSeq[0][0]) is float :
            for i in range(len(stimSeq)):
                for j in range(len(stimSeq[i])):
                    v = int(stimSeq[i][j]*scale)
                    if minval is not None: v=max(minval,v)
                    if maxval is not None: v=min(minval,v)
                    stimSeq[i][j] = v
        return stimSeq

    def setStimRate(self,rate):
        '''rewrite the stimtimes to equal given rate in hz'''
        setStimRate(self.stimTime_ms,rate)
        
    @staticmethod
    def readArray(f,width=-1):
        array=[]
        nEmpty=0
        for line in f:
            line = line.strip();
            if len(line)==0 :
                nEmpty += 1
                if nEmpty>1 and len(array)>0 : break # double empty means end-of-array
                else: continue 
            elif line[0]=="#" : continue # comment line
            cols = line.split();
            if width<0 : width=len(line)
            elif width>0 and not len(cols) == width : 
                raise Exception
            cols = [ float(c) for c in cols ] # convert string to numeric
            array.append(cols) # add to the stimSeq
        return array

    @staticmethod
    def fromString(f):
        """read a stimulus-sequence definition from a string"""
        st=StimSeq.readArray(f) # read the stim times
        if len(st) > 1:
            raise Exception
        else:
            st = st[0]  # un-nest
        ss = StimSeq.readArray(f, len(st))  # read stim-seq - check same length
        # transpose ss to have time in the major dimension
        ss = transpose(ss)
        return StimSeq(st, ss)

    @staticmethod
    def fromFile(fname):
        """read a stimulus-sequence from a file on disk"""
        import os.path
        if not os.path.isfile(fname):
            pydir = os.path.dirname(os.path.abspath(__file__))
            print('pydir={}'.format(pydir))
            fname = os.path.join(pydir, fname) if os.path.isfile(os.path.join(pydir, fname)) else fname
        if '.png' in fname:
            try:
                # try to load from the png
                from matplotlib.pyplot import imread
                array = imread(fname)
                if array.ndim > 2:
                    array = array[:, :, :3].mean(-1)
                array = array.tolist()
                ss = StimSeq(None, array, None)
                return ss
            except:
                pass
        
        f = open(fname, 'r')
        ss = StimSeq.fromString(f)
        return ss

    def toFile(self, fname):
        if '.png' in fname:
            # write out as a .png file
            # convert to byte with range 0-255
            import numpy as np
            import matplotlib
            array = np.array(self.stimSeq)
            array = array * 255 / np.max(array.ravel())
            array = array.astype(np.uint8)
            array = np.tile(array[:,:,np.newaxis],(1,1,3)) # convert to rgb
            print("{}".format(array.shape))
            matplotlib.image.imsave(fname,array,dpi=1)

        else:
            with open(fname,'w') as f:
                # ensure has stimTimes
                if self.stimTime_ms is None:
                    self.stimTime_ms = list(range(len(self.stimSeq)))
                f.write(str(self))

def transpose(M):
    return [[row[i] for row in M] for i in range(len(M[0]))]

def setStimRate(stimTime_ms,framerate):
    # rewrite the stim-times to a new frequency
    for i in range(len(stimTime_ms)):
        stimTime_ms[i] = i*1000/framerate
    return stimTime_ms


def mkRowCol(width=5,height=5, repeats=10):
    """Make a row-column stimulus sequence

    Args:
        width (int, optional): width of the matrix. Defaults to 5.
        height (int, optional): height of the matrix. Defaults to 5.
        repeats (int, optional): number of random row->col repeats. Defaults to 10.

    Returns:
        [StimSeq]: The generated stimulus sequence
    """    
    import numpy as np
    array = np.zeros((repeats,width+height,width,height))
    for ri in range(repeats):
        ei=0
        for r in np.random.permutation(width):
            array[ri,ei,:,r]=1
            ei=ei+1
        for c in np.random.permutation(height):
            array[ri,ei,c,:]=1
            ei=ei+1
    array = array.reshape((repeats*(width+height),width*height)) # (e,(w*h)) = (nEvent,nStim)
    return StimSeq(None,array.tolist(),None)


def mkFreqTag(period_phase=((3,0),(4,0),(5,0),(6,0),(7,0),(3,1),(4,1),(5,1),(6,1),(7,1),(3,2),(4,2),(5,2),(6,2),(7,2),(4,3),(5,3),(6,3),(7,3),(5,4),(6,4),(7,4),(6,5),(7,5),(7,6)),nEvent=120, isbinary=True):
    """Generate a frequency tagging stimulus sequence

    Args:
        period_phase (list, optional): List of stimulus periods and phase offsets expressed in events. Defaults to all phases from 2-5 long.
        nEvent (int, optional): total length of the sequence to generate. (To avoid steps should be LCM of periods) Defaults to 120.
        isbinary (bool, optional): flag if we generate a binary sequence or continuous
    """
    import numpy as np
    array = np.zeros((nEvent,len(period_phase)))
    for i,e in enumerate(period_phase):
        # extract desired length and phase
        l,o = e if hasattr(e,'__iter__') else (e,0)
        # generate the sequence
        s = np.sin( (np.arange(array.shape[0])+o+1e-6)/l*2*np.pi )
        if isbinary:
            array[s>0,i] = 1
        else:
            array[:,i] = s
    return StimSeq(None,array.tolist(),None)


def mkCodes():
    # test generators
    rc=mkRowCol(width=5,height=5, repeats=10)
    rc.toFile('rc5x5.png')
    rc.toFile('rc5x5.txt')

    ssvep=mkFreqTag()
    ssvep.toFile('ssvep.png')
    ssvep.toFile('ssvep.txt')

# testcase code
if __name__ == "__main__":
    # loading text version
    print(("gold(txt): " + str(StimSeq.fromFile("mgold_65_6532_psk_60hz.txt"))))

    # loading png version
    print(("gold(png): " + str(StimSeq.fromFile("mgold_61_6521_psk_60hz.png"))))
    
    # test writing to file
    ss = StimSeq.fromFile("mgold_65_6532_psk_60hz.txt")
    ss.toFile("mgold_65_6532_psk_60hz.png")

    # make codebooks
    mkCodes()


import numpy as np
import matplotlib.pyplot as plt
from mindaffectBCI.decoder.datasets import get_dataset
from mindaffectBCI.decoder.model_fitting import BaseSequence2Sequence, MultiCCA, FwdLinearRegression, BwdLinearRegression, LinearSklearn
from sklearn.linear_model import Ridge, LogisticRegression
from sklearn.svm import LinearSVR, LinearSVC
from sklearn.model_selection import cross_val_score
from mindaffectBCI.decoder.updateSummaryStatistics import updateSummaryStatistics, plot_erp, plot_summary_statistics
from mindaffectBCI.decoder.scoreStimulus import factored2full, plot_Fe
from mindaffectBCI.decoder.decodingCurveSupervised import decodingCurveSupervised, print_decoding_curve, plot_decoding_curve
from mindaffectBCI.decoder.utils import lab2ind, butter_sosfilt, extract_envelope
from mindaffectBCI.decoder.scoreOutput import plot_Fy
from mindaffectBCI.decoder.multipleCCA import robust_whitener
from mindaffectBCI.decoder.updateSummaryStatistics import updateCxx, idOutliers
from scipy.signal import welch

l,f,_=get_dataset('openBMI_MI')

# get raw MI data, just for the trial duration
#X,Y,coords=l(f[1],stopband=((0,3),(29,-1)),CAR=False,offset_ms=None,ppMI=False)
offset_ms=(0,0);(-6000,12000)
stopband=None #((0,3),(30,-1))
X,Y,coords=l(f[0],stopband=stopband,CAR=True,offset_ms=offset_ms,ppMI=False,fs_out=100)
oX=X.copy()
fs=coords[1]['fs']
lab=coords[0]['lab']

def plot_erp(X,lab,title,fs=None,chnames=None,plotp=True,differp=False):
    lab,key=lab2ind(lab)
    plt.clf();
    plt.subplot(211);
    plt.imshow(X[0,:,:].copy(),aspect='auto');
    erp=np.einsum("Tsd,Ty->ysd",X,lab);
    if differp:
        erp = erp - np.mean(erp,0,keepdims=True) # center erps, so show diffs
    for yi in range(erp.shape[0]):
        plt.subplot(2,erp.shape[0],erp.shape[0]+yi+1);
        if plotp:
            plt.plot(erp[yi,:,:]);
            plt.grid(True)
        else:
            plt.imshow(erp[yi,:,:],aspect='auto');
        plt.xlabel('space (ch*freq)')
        plt.title(key[yi])
    plt.suptitle(title)

X=oX.copy()
plt.figure(100);plot_erp(X,lab,'raw')

# CAR
X = X - np.mean(X,-1,keepdims=True)

# outlier removal
badtr,pow = idOutliers(X,axis=(1,2))
print('Removing {} bad trials'.format(np.sum(badtr.ravel())))
keep = badtr.ravel()==False
X = X[keep,...]
lab=lab[keep]

badch,_ = idOutliers(X,axis=(0,1))
print('Removing {} bad channels'.format(np.sum(badch.ravel())))
keep = badch.ravel()==False
X = X[..., keep]
ch_names=[ch_names[i] for i in range(len(ch_names)) if keep[i]]

# ch-subset
gigasubset=('C5','C3','C1','Cz','C2','C4','C6',\
            'FC5','FC3','FC1','FCz','FC2','FC4','FC6',\
            'CP5','CP3','CP1','CPz','CP2','CP4','CP6')
minsubset =('C3','C4')
chsubset=minsubset
keep=[c in chsubset for c in ch_names]
X = X[...,keep]
ch_names = [ ch_names[i] for i in range(len(ch_names)) if keep[i] ]
plt.figure(100);plot_erp(X,lab,'car')

# hp-lp
X,_,_=butter_sosfilt(X,stopband=((0,8),(16,-1)),fs=fs)
plt.figure(101);plot_erp(X,lab,'hp-lp',plotp=True)

# whiten
Cxx=updateCxx(None,X,None)
W,_=robust_whitener(Cxx)
X  =np.einsum("Tsd,dw->Tsw",X,W)
#Cxxw=updateCxx(None,X,None)

plt.figure(102);plot_erp(X,lab,'wht')

# welch
freqs, X = welch(X,fs,axis=-2,nperseg=int(fs*.5),noverlap=.5,return_onesided=True,scaling='spectrum')
plt.figure(103);plot_erp(X,lab,'welch')

# envelope
X = extract_envelope(X,fs,stopband=None,whiten=None,filterbank=None,log=False,env_stopband=(2,-1),plot=False)
plt.figure(103);plot_erp(X,lab,'env',plotp=True)

X=np.log(np.maximum(X,1e-5))

# baseline
bl_ms=(-1000,0)
bl_idx = slice(int((bl_ms[0]-offset_ms[0])*fs/1000),int((bl_ms[1]-offset_ms[0])*fs/1000))
X = X - np.mean(X[:,bl_idx,:],axis=1,keepdims=True)
plt.figure(103);plot_erp(X,lab,'env',plotp=True)

# classify
clsfr=LogisticRegression(C=1e8)
cross_val_score(clsfr,X.reshape((X.shape[0],-1)),lab,cv=5)



import numpy as np
from mindaffectBCI.utopiaclient import StimulusEvent
def devent2stimSequence(devents):
    '''
    convert a set of STIMULUSEVENT messages into a stimulus-sequence array with stimulus-times as expected by the utopia RECOGNISER
    
    Inputs:
     devents - [nEvt]:UtopiaMessage list of UtopiaMessage messages
            i.e. a decoded utopia STIMULUSEVENT message, should be of type
             { msgID:'E'byte, timeStamp:int, objIDs:(nObj:byte), objState:(nObj:byte) }
    Outputs:
     Me     - (nEvt,nY :int) The extract stimulus sequence
     objIDs - (nY :byte) The set of used object IDs
     stimTimes_ms - (nEvt :int) The timestamp of each event in milliseconds
     isistimEvent - (nEp :bool) Indicator which input events are stimulus events
      
    Copyright (c) MindAffect B.V. 2018
    '''
    if devents is None:
        return (None,None,None,None)
    Me = np.zeros((len(devents), 256), dtype=int)
    stimTimes_ms = np.zeros(len(devents))
    allobjIDs = np.arange(0, 256)
    usedobjIDs = np.zeros((256), dtype=bool)
    isstimEvent = np.zeros((len(devents)), dtype=bool)

    # loop over the messages, extracting info from stimEvents and putting into the stim-sequence
    for ei, evt in enumerate(devents):
        if not evt.msgID == StimulusEvent.msgID:
            continue
        isstimEvent[ei] = True
        # extract the stimulusState info
        timeStamp = evt.timestamp
        objIDs = evt.objIDs
        objState = evt.objState
        stimTimes_ms[ei] = timeStamp
        # hold value of used objIDs from previous time stamp
        if ei > 0 and np.any(usedobjIDs):
            Me[ei, usedobjIDs] = Me[ei-1, usedobjIDs]
        # 2) overwrite with updated state
        # N.B. indexing hack with [int,tuple,] to index correctly
        Me[ei, objIDs, ] = objState
        usedobjIDs[objIDs, ] = True
    
    # restrict to only stim-event info
    # Note: horrible integer indexing tricks to get only the elements we want..
    Me = Me[np.flatnonzero(isstimEvent)[:, np.newaxis], usedobjIDs]
    stimTimes_ms = stimTimes_ms[isstimEvent]
    objIDs = allobjIDs[usedobjIDs]
    return Me, stimTimes_ms, objIDs, isstimEvent



def upsample_stimseq(sample_ts, ss, stimulus_ts, objIDs=None, usedobjIDs=None, trlen=None):
    ''' upsample a set of timestamped stimulus states to a sample rate
     WARNING: assumes sample_ts and stimulus_ts are in *ascending* sorted order! '''
    if trlen is None:
        trlen = len(sample_ts)

    if objIDs is None:
        objIDs = list(range(ss.shape[-1]))

    if usedobjIDs is None: # use the whole Y
        usedobjIDs = objIDs
        obj_idx = slice(len(usedobjIDs))

    else: # match objIDs and to get idxs to use
        obj_idx = np.zeros((len(objIDs),), dtype=int)
        for oi, oid in enumerate(objIDs):
            obj_idx[oi] = np.flatnonzero(usedobjIDs == oid)[0]
   
    # convert to sample-rate version
    Y = np.zeros((trlen, len(usedobjIDs)), dtype=ss.dtype)
    samp_ts_iterator = enumerate(iter(sample_ts))
    data_i=None
    stimulus_idx=np.zeros(stimulus_ts.shape,dtype=int)
    for stim_i, stim_ts in enumerate(stimulus_ts):
        #print("{}) ts={} ".format(stim_i,stim_ts),end='')
        odata_i = data_i # store when old stim sample index
        for data_i, data_ts in samp_ts_iterator:
            #print("{}={} ".format(data_i,data_ts),end='')
            # N.B. assumes samp_ts are already interpolated!
            if data_ts > stim_ts:
                #print("*")
                break
        if data_i is None:
            raise ValueError("Ran out of data!")
        if data_i > Y.shape[0]:
            # events after end of the allowed trial length
            break
        # data_i is one sample too far?
        data_i = max(data_i-1, 0)
        # nearest index for the stim_ts
        if odata_i is not None:
            # hold the previous stimulus state until now
            Y[odata_i:data_i+1, obj_idx] = Y[odata_i, obj_idx]
        # insert the new state
        stimulus_idx[stim_i] = data_i
        Y[data_i, obj_idx] = ss[stim_i, :]
                
    return (Y, stimulus_idx)


if __name__=="__main__":
    from devent2stimsequence import devent2stimSequence
    from mindaffectBCI.utopiaclient import StimulusEvent
    # make a periodic type stimulus sequence, period, 3,4,5
    se = [StimulusEvent(i*3, (1, 2, 3),
                        (i%3 == 0, i%4 == 0, i%5 == 0)) for i in range(100)]
    Me, st, oid, isse = devent2stimSequence(se)

    # print the decoded sequence
    print("oID  {}".format(oid))
    print("---")
    for i in range(Me.shape[0]):
        print("{:3.0f}) {}".format(st[i], Me[i, :]))

    # now check the upsampling function, where we sampled at 1hz
    used_objIDs = np.arange(5)
    samp_ts = np.arange(len(se)*5)
    trlen = int(st[-1]*.6) # make trial shorter than data
    Y,_ = upsample_stimseq(samp_ts, Me, st, oid, used_objIDs, trlen)
    print("oID  {}".format(used_objIDs))
    print("\n---\n")
    for i in range(Y.shape[0]):
        print("{:3d}) {}".format(i,Y[i,:]))
        



from bluepy.btle import DefaultDelegate, Peripheral, Scanner
import atexit
import sys
import warnings
import datetime
from bitstring import BitArray
import numpy as np


# TODO: Add aux data
# TODO: Reconnecting when dropped

SAMPLE_RATE = 200.0  # Hz

# service for communication, as per docs
BLE_SERVICE = "fe84"
# characteristics of interest
BLE_CHAR_RECEIVE = "2d30c082f39f4ce6923f3484ea480596"
BLE_CHAR_SEND = "2d30c083f39f4ce6923f3484ea480596"
BLE_CHAR_DISCONNECT = "2d30c084f39f4ce6923f3484ea480596"

class OpenBCIGanglion(object):
    """ OpenBCIGanglion handles the connection to an OpenBCI Ganglion board.

    The OpenBCIGanglion class interfaces with the Cyton Dongle and the Cyton board to parse the data received and output it to Python as a OpenBCISample object.

    Args:
        mac: A string representing the Ganglion board mac address. It should be a string comprising six hex bytes separated by colons, e.g. "11:22:33:ab:cd:ed". If no mac address specified, a connection will be stablished with the first Ganglion found (Will need root privilages).

        max_packets_skipped: An integer specifying how many packets can be dropped before attempting to reconnect.
    """
    def __init__(self, mac=None,sample_rate=None, max_packets_skipped=15):
        if not mac:
            self.mac_address = self.find_mac()
        else:
            self.mac_address = mac
        self.max_packets_skipped = max_packets_skipped
        self.streaming = False
        self.board_type = 'Ganglion'
        self.sample_rate = sample_rate

        atexit.register(self.disconnect)

        self.connect()

    def write_command(self, command):
        """Sends string command to the Ganglion board."""
        self.char_write.write(str.encode(command))

    def connect(self):
        """Establishes connection with the specified Ganglion board."""
        self.ganglion = Peripheral(self.mac_address, 'random')

        self.service = self.ganglion.getServiceByUUID(BLE_SERVICE)

        self.char_read = self.service.getCharacteristics(BLE_CHAR_RECEIVE)[0]

        self.char_write = self.service.getCharacteristics(BLE_CHAR_SEND)[0]

        self.char_discon = self.service.getCharacteristics(BLE_CHAR_DISCONNECT)[0]

        self.ble_delegate = GanglionDelegate(self.max_packets_skipped)
        self.ganglion.setDelegate(self.ble_delegate)

        self.desc_notify = self.char_read.getDescriptors(forUUID=0x2902)[0]

        try:
            self.desc_notify.write(b"\x01")
        except Exception as e:
            print("Something went wrong while trying to enable notification: " + str(e))
            sys.exit(2)

        print("Connection established")

    def disconnect(self):
        """Disconnets from the Ganglion board."""
        if self.streaming:
            self.stop_stream()

        self.char_discon.write(b' ')
        self.ganglion.disconnect()

    def find_mac(self):
        """Finds and returns the mac address of the first Ganglion board found"""
        scanner = Scanner()
        devices = scanner.scan(5)

        if len(devices) < 1:
            raise OSError('No nearby Devices found. Make sure your Bluetooth Connection is on.')

        else:
            gang_macs = []
            for dev in devices:
                for adtype, desc, value in dev.getScanData():
                    if desc == 'Complete Local Name' and value.startswith('Ganglion'):
                        gang_macs.append(dev.addr)
                        print(value)

        if len(gang_macs) < 1:
            raise OSError('Cannot find OpenBCI Ganglion Mac address.')
        else:
            print("Connecting to Ganglion with mac address: "+ gang_macs[0])
            return gang_macs[0]

    def stop_stream(self):
        """Stops Ganglion Stream."""
        self.streaming = False
        self.write_command('s')

    def start_stream(self, callback):
        """Start handling streaming data from the Ganglion board. Call a provided callback for every single sample that is processed."""
        if not self.streaming:
            self.streaming = True
            self.dropped_packets = 0
            self.write_command('b')

        if not isinstance(callback, list):
            callback = [callback]

        while self.streaming:
            try:
                self.ganglion.waitForNotifications(1./SAMPLE_RATE)
            except Exception as e:
                print(e)
                print('Something went wrong')
                sys.exit(1)

            samples = self.ble_delegate.getSamples()
            if samples:                                
                #for sample in samples:
                for call in callback:
                    call(samples) # call with block samples in 1 go
                        
    def getSampleRate(self):
        return self.sample_rate


class GanglionDelegate(DefaultDelegate):
    """ Delegate Object used by bluepy. Parses the Ganglion Data to return an OpenBCISample object.
    """
    def __init__(self, max_packets_skipped=15):

        DefaultDelegate.__init__(self)
        self.max_packets_skipped = max_packets_skipped
        self.last_values = [0, 0, 0, 0]
        self.last_id = -1
        self.samples = []
        self.start_time = datetime.datetime.now().strftime("%Y-%m-%d_%H%M%S")

    def handleNotification(self, cHandle, data):
        """Called when data is received. It parses the raw data from the Ganglion and returns an OpenBCISample object"""
        if len(data) < 1:
            warnings.warn('A packet should at least hold one byte...')
        self.parse_raw(data)

    def parse_raw(self, raw_data):
        """Parses the data from the Cyton board into an OpenBCISample object."""
        #if type(raw_data) == str:
        #    data = struct.unpack(str(len(packet)) + 'B', "".join(packet))
        #else:
        #    data = raw_data

        start_byte = raw_data[0]
        bit_array = BitArray()
        self.checked_dropped(start_byte)
        # print(start_byte, start_byte == 0)

        if start_byte == 0:
            # 24-bit encoded
            for byte in raw_data[1:13]:
                bit_array.append('0b{0:08b}'.format(byte))
            results = []
            # and split it into 24-bit chunks here
            for sub_array in bit_array.cut(24):
                # calling ".int" interprets the value as signed 2's complement
                results.append(sub_array.int)
                self.last_values = np.array(results)
                # print(self.last_values)
                self.push_sample( [np.append(start_byte, self.last_values)])

        elif start_byte >=1 and start_byte <=100:
            # 18-bit encoded + accel
            for byte in raw_data[1:-1]:
                bit_array.append('0b{0:08b}'.format(byte))
            deltas = []
            for sub_array in bit_array.cut(18):
                deltas.append(self.decompress_signed(sub_array))

            delta1 , delta2 = np.array(deltas[:4]) , np.array(deltas[4:])
                
            self.last_values1 = self.last_values - delta1
            self.last_values = self.last_values1 - delta2
                
            self.push_sample([self.last_values1, self.last_values])

        elif start_byte >=101 and start_byte <=200:
            # 19-bit encoded
            for byte in raw_data[1:]:
                bit_array.append('0b{0:08b}'.format(byte))
            deltas = []
            for sub_array in bit_array.cut(19):
                deltas.append(self.decompress_signed(sub_array))

            delta1 , delta2 = np.array(deltas[:4]) , np.array(deltas[4:])
            self.last_values1 = self.last_values - delta1
            # print(self.last_values1)
            self.last_values = self.last_values1 - delta2
            # print(self.last_values1)
            self.last_values = self.last_values1 - delta2
            # print(self.last_values)
                
            self.push_sample([np.append(start_byte,self.last_values1), np.append(start_byte,self.last_values)])

        # self.push_sample(data)

    def push_sample(self, data):
        """Creates a stack with the last ganglion Samples"""
        for data_arr in data:
            if len(data_arr) == 5:
                sample = OpenBCISample(data_arr[0], data_arr[1:], [], self.start_time, 'Ganglion')
                self.samples.append(sample)

    def getSamples(self):
        """Returns the last OpenBCI Samples in the stack"""
        old_samples = self.samples
        self.samples = []
        return old_samples

    def checked_dropped(self, num):
        """Checks dropped packets"""
        if num not in [206, 207]:
            if self.last_id == 0 and num not in [1, 101]:
                if num > 100:
                    dropped = num - 100
                else:
                    dropped = num
            elif self.last_id == 0:
                dropped = 0
            elif self.last_id > num:
                dropped = 100 - abs(self.last_id - num)
            else:
                dropped = abs(self.last_id - num)

            if dropped > 1:#self.max_packets_skipped:
                print("Dropped %d packets...." % dropped)

            self.last_id = num

    def decompress_signed(self, bit_array):
        """Used to decrompress signed bit arrays."""
        result = bit_array.int
        if bit_array.endswith(1):#'0b1'):
            result -= 1
        return result

class OpenBCISample():
    """ Object that encapsulates a single sample from the OpenBCI board.

    Attributes:
        id: An int representing the packet id of the aquired sample.
        channels_data: An array with the data from the board channels.
        aux_data: An array with the aux data from the board.
        start_time: A string with the stream start time.
        board_type: A string specifying the board type, e.g 'cyton', 'daisy', 'ganglion'
    """
    def __init__(self, packet_id, channels_data, aux_data, init_time, board_type):
        self.id = packet_id
        self.channels_data = channels_data
        self.aux_data = aux_data
        self.start_time = init_time
        self.board_type = board_type



import struct
import numpy as np
import os
import array

typedict= {0:('char',1,'c'),
           1:('uint8',1,'B'),
           2:('uint16',2,'H'),
           3:('uint32',4,'I'),
           4:('unint64',8,'Q'),
           5:('int8',1,'b'),
           6:('int16',2,'h'),
           7:('int32',4,'i'),
           8:('int64',8,'l'),
           9:('float32',4,'f'),
           10:('float64',8,'d')}

class ftevent:
    def __init__(self,s,t,v,o=None,d=None):
        self.sample=s
        self.type=t
        self.value=v
        self.offset=o
        self.duration=d
    def __str__(self):
        return "{{ s:{} t:{} v:{} }}".format(self.sample,self.type,self.value)

class ftheader:
    def __init__(self,nch,nsamp,nevt,fs,data_type,labels=None):
        self.nch=nch
        self.nsamp=nsamp
        self.nevt=nevt
        self.fs=fs
        self.data_type=typedict[data_type] if isinstance(data_type,str) else data_type
        self.labels=labels

    
def read_buffer_offline_events(f):
    if isinstance(f,str):
        if os.path.isdir(f):
            f = os.path.join(f,'events')
        buf = open(f,'rb').read()
    else:
        buf = f

    devents=[]
    cursor=0
    while cursor<len(buf):
        t_type,t_len,v_type,v_len,sample,offset,duration,payloadsize = struct.unpack_from("IIIIiiiI",buf,cursor)
        cursor+=struct.calcsize("IIIIiiiI")
        payload = buf[cursor:cursor+payloadsize]
        cursor+=payloadsize

        # decode the payload
        tfmtstr = "{}{}".format(t_len,typedict[t_type][2])
        vfmtstr = "{}{}".format(v_len,typedict[v_type][2])
        if struct.calcsize(tfmtstr)+struct.calcsize(vfmtstr) > len(payload):
            print("Error: didn't understand event info, skipped")
            continue

        
        if typedict[t_type][2] == 'c':
            tval = payload[:struct.calcsize(tfmtstr)].decode('utf8')
        else:
            tval=struct.unpack_from(tfmtstr,payload,0)
        if typedict[v_type][2] == 'c':
            vval = payload[struct.calcsize(tfmtstr):].decode('utf8')
        else:
            vval=struct.unpack_from(vfmtstr,payload,struct.calcsize(tfmtstr))
        # make the event
        devents.append(ftevent(sample,tval,vval,offset,duration))
    return devents

def read_buffer_offline_header(f):
    if isinstance(f,str):
        if os.path.isdir(f):
            f = os.path.join(f,'header')
        buf = open(f,'rb').read()
    else:
        buf = f
    nch,nsamp,nevts,fs,data_type,bufsize=struct.unpack_from("IIIfII",buf,0)
    # TODO []: extract the channel labels
    labbuf = buf[struct.calcsize("IIIfII"):]
    # TODO []: read the channel names...
    return ftheader(nch,nsamp,nevts,fs,data_type,None)

def read_buffer_offline_data(f,hdr):
    fmtstr = hdr.data_type if isinstance(hdr.data_type,str) else typedict[hdr.data_type][2]
    if isinstance(f,str):
        if os.path.isdir(f):
            f = os.path.join(f,'samples')
        dat = np.fromfile(f,dtype=typedict[hdr.data_type][0],count=-1)
    else:
        dat = dat.frombuffer(buf,dtype=typedict[hdr.data_type][0],count=-1)
    dat = np.reshape(dat,(len(dat)//hdr.nch,hdr.nch))
    return dat

def testcase():
    from read_buffer_offline import read_buffer_offline_data, read_buffer_offline_events, read_buffer_offline_header
    datadir="../../resources/example_data/utopia_v1/s1/raw_buffer/0001"
    hdr = read_buffer_offline_header(datadir)
    print("{}".format(hdr))
    evts= read_buffer_offline_events(datadir)
    print("{} events".format(len(evts)))
    dat = read_buffer_offline_data(datadir,hdr)
    print("dat={}".format(dat.shape))    


if __name__=='__main__':
    testcase()



import os
import numpy as np
from scipy.io import loadmat
from mindaffectBCI.decoder.utils import butter_sosfilt, window_axis, block_randomize

def load_p300_prn(datadir, sessdir=None, sessfn=None, fs_out=60, offset_ms=(-1000,1000), ifs=None, fr=None, stopband=((45,65), (0,1), (25,-1)), order=4, subtriallen=10, verb=0, nvirt=20, chidx=slice(64)):

    # load the data file
    Xfn = datadir
    if sessdir:
        Xfn = os.path.join(Xfn, sessdir)
    if sessfn:
        Xfn = os.path.join(Xfn, sessfn)
    sessdir = os.path.dirname(Xfn)
    try:
        data = loadmat(Xfn)
    except NotImplementedError:
        # TODO[] : make this work correctly -- HDF5 field access is different to loadmat's
        #import h5py
        #data = h5py.File(Xfn, 'r')
        pass

    X = data['X']
    X = X.astype("float32") # [ ch x samp x trl ]:float - raw eeg
    X = np.moveaxis(X, (0, 1, 2), (2, 1, 0)) # (nTrl, nSamp, d)
    X = np.ascontiguousarray(X) # ensure memory efficient layout

    ch_names = np.stack(data['di'][0]['vals'][0][0]).ravel()
    # Extract the sample rate.  Argh!, why so deeply nested?
    if ifs is None:
        fs = data['di'][1]['info'][0]['fs'][0,0][0,0]
    else:
        fs = ifs

    extrainfo = data['di'][2]['extra'][0]
    try:
        Ye = np.stack(extrainfo['flipgrid'][0], -1) # (nY,nEp,nTrl)
    except:
        Ye = None
    Ye0= np.stack(extrainfo['flash'][0], -1) # true-target  (1,nEp,nTrl)
    tgtLetter = extrainfo['target']  # target letter, not needed

    samptimes = data['di'][1]['vals'][0].ravel()   # (nSamp,)
    flashi_ms = np.stack(extrainfo['flashi_ms'][0], -1) #(1,nEp,nTrl)

    # convert flashi_ms to flashi_samp and upsampled Ye to sample rate
    Ye0= np.moveaxis(Ye0, (0, 1, 2), (2, 1, 0)) # (nTrl, nEp, 1)
    stimTimes_ms = np.moveaxis(flashi_ms, (0, 1, 2), (2, 1, 0)) # (nTrl, nEp, 1)
    if Ye is not None:
        Ye = np.moveaxis(Ye,  (0, 1, 2), (2, 1, 0)) # (nTrl, nEp, nY)
    else:
        # make a pseudo-set of alternative targets
        Ye = block_randomize(Ye0[...,np.newaxis],nvirt,-3) #(nTrl,nEp,nvirt,1)
        Ye = Ye[...,0] # (nTrl,nEp,nvirt)
        print("{} virt targets".format(Ye.shape[-1]))
    
    # upsample to sample rate
    stimTimes_samp = np.zeros(stimTimes_ms.shape, dtype=int) # index from trial start for each flash
    Y = np.zeros(X.shape[:-1]+(Ye.shape[-1]+Ye0.shape[-1],), dtype='float32') # (nTrl, nEP, nY+1)
    for ti in range(Y.shape[0]):
        lastflash = None
        flashi_trli = stimTimes_ms[ti, :, 0]
        for fi, flash_time_ms in enumerate(flashi_trli):
            # find nearest sample time
            si = np.argmin(np.abs(samptimes - flash_time_ms))
            stimTimes_samp[ti, fi, 0] = si
            
            if lastflash: # hold until new values
                Y[ti, lastflash+1:si, :] = Y[ti, lastflash, :]
                
            Y[ti, si, 0]  = Ye0[ti, fi, 0] # true info always 1st row
            Y[ti, si, 1:] = Ye[ti, fi, :]  # rest possiblities
            lastflash = si
    # for comparsion...
    #print("{}".format(np.array(np.mean(stimTimes_samp, axis=0),dtype=int).ravel()))
    
    # preprocess -> ch-seln
    if chidx is not None:
        X=X[...,chidx]
        ch_names = ch_names[chidx]

    # Trim to useful data range
    stimRng = ( np.min(stimTimes_samp[:,0,0]+offset_ms[0]*fs/1000),
                np.max(stimTimes_samp[:,-1,0]+offset_ms[1]*fs/1000) )
    print("stimRng={}".format(stimRng))
    if  0 < stimRng[0] or stimRng[1] < X.shape[-2]:
        if verb>-1 : print('Trimming range: {}-{}ms'.format(stimRng[0]/fs,stimRng[-1]/fs))
        # limit to the useful data range
        rng = slice(int(max(0,stimRng[0])), int(min(X.shape[-2],stimRng[1])))
        X = X[..., rng, :]
        Y = Y[..., rng, :]
        if verb > 0: print("X={}".format(X.shape))
        if verb > 0: print("Y={}".format(Y.shape))

    # preprocess -> spectral filter
    if stopband is not None:
        if verb > 0:
            print("preFilter: {}Hz".format(stopband))
        X, _, _ = butter_sosfilt(X,stopband,fs,order=order)
    
    # preprocess -> downsample 
    resamprate = int(round(fs/fs_out))
    if resamprate > 1:
        if verb > 0:
            print("resample: {}->{}hz rsrate={}".format(fs, fs_out, resamprate))
        X = X[:, ::resamprate, :] # decimate X (trl, samp, d)
        Y = Y[:, ::resamprate, :] # decimate Y (trl, samp, y)
        fs = fs/resamprate

    nsubtrials = X.shape[1]/fs/subtriallen if subtriallen is not None else 0
    if nsubtrials > 1:
        winsz = int(X.shape[1]//nsubtrials)
        print('{} subtrials -> winsz={}'.format(nsubtrials,winsz))
        # slice into sub-trials
        X = window_axis(X,axis=1,winsz=winsz,step=winsz) # (trl,win,samp,d)
        Y = window_axis(Y,axis=1,winsz=winsz,step=winsz) # (trl,win,samp,nY)
        # concatenate windows into trial dim
        X = X.reshape((X.shape[0]*X.shape[1],)+X.shape[2:])
        Y = Y.reshape((Y.shape[0]*Y.shape[1],)+Y.shape[2:])
        if verb>0 : print("X={}".format(X.shape))

    # make coords array for the meta-info about the dimensions of X
    coords = [None]*X.ndim
    coords[0] = {'name':'trial','coords':np.arange(X.shape[0])}
    coords[1] = {'name':'time','unit':'ms', \
                 'coords':np.arange(X.shape[1]) * 1000/fs, \
                 'fs':fs}
    coords[2] = {'name':'channel','coords':ch_names}
    
    return (X, Y, coords)

def testcase():
    import sys
    from readCapInf import getPosInfo

    datadir = '/home/jadref/data/bci/own_experiments/visual/p300_prn_2'
    sessdir = ''
    sessfn = 'peter/20100714/jf_prep/peter_rc_5_flash.mat'
    sessfn = 'peter/20100714/jf_prep/peter_prn_5_flip.mat'

    #sessfn = 'alex/20100722/jf_prep/alex_prn_10_flip.mat'
    
    from offline.load_p300_prn import load_p300_prn
    X, Y, coords = load_p300_prn(datadir, sessdir, sessfn, fs_out=32, stopband=((0,1),(12,-1)), order=6); oX=X.copy(); oY=Y.copy();
    fs = coords[1]['fs']
    ch_names = coords[2]['coords']

    # attach the electrode positions
    cnames, xy, xyz, iseeg =getPosInfo(ch_names)
    coords[2]['pos2d'] = xy
    
    # set the analysis paramters
    tau = int(.6*fs)
    evtlabs = ('re','anyre') # ('re', 'ntre') #('re', 'fe') # 
    times = [i/fs for i in range(tau)]
    rank = 10

    from analyse_datasets import debug_test_dataset
    debug_test_dataset(X, Y, coords, tau_ms=800, evtlabs=evtlabs, rank=rank, regy=.1, regx=.1)
    

    # try with cross-validation
    res = cca.cv_fit(X, Y)
    Fy=res['estimator']
    (_) = decodingCurveSupervised(Fy)
    
    print('FWD')
    fwd = FwdLinearRegression(tau=tau, evtlabs=evtlabs)
    fwd.fit(X, Y)
    print('fwd = {}'.format(fwd))
    Fy = fwd.predict(X, Y, dedup0=True)
    print("Fy={}".format(Fy.shape))
    (_) = decodingCurveSupervised(Fy)

    plot_erp(factored2full(fwd.W_, fwd.R_), evtlabs, times, ch_names)

    
    print('BWD')
    bwd = BwdLinearRegression(tau=tau, evtlabs=evtlabs)
    bwd.fit(X, Y)
    print('bwd = {}'.format(bwd))
    Fy = bwd.predict(X, Y, dedup0=True)
    print("Fy={}".format(Fy.shape))
    (_) = decodingCurveSupervised(Fy)


if __name__ == "__main__":
    testcase()



import numpy as np
def normalizeOutputScores_streamed(Fy, validTgt=None, badFyThresh=6,
                                   centFy=True, blockSize=30, filtLen=5):
    '''
    normalize the raw output scores to feed into the Perr computation
    Inputs:
      Fy   - (nM,nTr,nEp,nY):float
      validTgt - (nM,nTr,nY):bool indication of which outputs are used in each trial
          badFyThresh - threshold for detction of bad Fy entry, in std-dev
          centFy   - bool, do we center Fy before computing  *important*  (true)
          blockSize  - maximum number of epochs single normalized sum
    Outputs:
      ssFy - (nM,nTr,nDecis,nY):float  scaled summed scores
      scale_sFy - (nM,nTr,nDecis):float scaling factor for each decision point
      N - (nDecis):int the number of elements summed in each block
    Copyright (c) MindAffect B.V. 2018
    
    '''
    normSum = True
    if Fy is None:
        return None, None, None, None, None

    # compress out the model dimension
    Fyshape = Fy.shape # (nM,nTrl,nEp,nY)
    if Fy.ndim > 3:
        Fy = np.reshape(Fy, (np.prod(Fy.shape[:-2]), )+Fy.shape[-2:]) # ((nM*nTrl), nEp, nY)
        if not validTgt is None : # (nM, nTr, nY)
            validTgt = np.reshape(validTgt, (np.prod(validTgt.shape[:-1]), ) + validTgt.shape[-1:]) # ((nM*nTrl), nY)
    if Fy.ndim < 3: # ensure has trial dim
        Fy = Fy[np.newaxis, :, :]

    # get info on the number of valid Epochs or Outputs in each trial
    if validTgt is None : # (nM, nTr, nEp, nY)
        validTgt = np.any(Fy != 0, -2) # (nM, nTr, nY) [ nY x nTrl ]

    # loop over trials
    nBlk = Fy.shape[1]//blockSize
    ssFy = np.zeros((Fy.shape[0], nBlk, Fy.shape[2]), dtype='float32')
    scale_sFy = np.zeros((Fy.shape[0], nBlk), dtype='float32')
    N = np.minimum(np.arange(1, nBlk+1)*blockSize, Fy.shape[1])
    for ti in range(Fy.shape[-3]):
        # N.B. sub-slice in steps as mixing integer with logical idexing is ackward
        Fyti = Fy[ti,...]
        # get num active epochs
        tmp = np.flatnonzero(np.any(Fy[ti, :, :] != 0, axis=-1))
        nEp = tmp[-1] if tmp.size > 0 else 0 # num active epochs/samples
        # limit to valid epochs
        Fyti = Fyti[:nEp, :]
        # limit to valid outputs
        validY = validTgt[ti, :]
        Fyti = Fyti[..., validY]
        
        # get the  start/end of each block
        decisIdx = np.arange(0, nEp-(blockSize//2), blockSize)
        decisIdx = np.append(decisIdx, nEp)

        # compute the noise-scale *incrementally* for each block
        sigma2 = None
        for bi in range(len(decisIdx)-1):
            Fytibi = Fyti[decisIdx[bi]:decisIdx[bi+1], ...]
            sigma2 = incremental_estimate_noise_variance(Fytibi, sigma2, filtLen)
            
            # apply the normalization and compute the summed score
            scale_sFy[ti,bi] = np.sqrt(Fytibi.shape[0]*sigma2)
            ssFy[ti,bi,validY] = np.sum(Fytibi,0) / scale_sFy[ti,bi]            

    if len(Fyshape)>3 : # convert back to have model dimension
        ssFy      = np.reshape(ssFy, (Fyshape[:-2]+ssFy.shape[-2:]))

    return (ssFy, scale_sFy, N)


def incremental_estimate_noise_variance(Fy, old_sigma2, filtLen):
    # TODO []: pre-filter Fy?
    cFy = Fy - np.mean(Fy, -1, keepdims=True) # per-output offset centered
    cFy = cFy - np.mean(cFy, 0, keepdims=True) # per-time point offset centered
    sigma2 = np.sum(cFy**2)/cFy.size # average residual
    # accumulate over calls
    if old_sigma2:
        sigma2 = old_sigma2*.5 + .5*sigma2
    return sigma2


def testcase(Fy=None):
    from normalizeOutputScores import mktestFy,  normalizeOutputScores, plot_normalizedScores
    from normalizeOutputScores_streamed import normalizeOutputScores_streamed, incremental_estimate_noise_variance
    from zscore2Ptgt_softmax import softmax
    import matplotlib.pyplot as plt
    if Fy is None:
        Fy, _ = mktestFy() #(nM, nTrl, nEp, nY)
        Fy = Fy[0, ...] # strip model dim
    print("Fy={}".format(Fy.shape))
    # visualize all trials true-target normalized scores
    blk_len = 15
    zFy, scale_sFy, N = normalizeOutputScores_streamed(Fy, blockSize=blk_len)
    print('zFy={}'.format(zFy.shape))
    sumcorr = np.sqrt(np.arange(zFy.shape[-2])+1) # correction for block sums
    plt.figure(2);plt.clf()
    plot_normalizedScores(Fy[0,:,:],np.cumsum(zFy[0,:,:],-2)/sumcorr[:,np.newaxis],np.cumsum(scale_sFy[0,:],-1)/sumcorr,N)

    # compare: sum->pVal with pval->combine
    # sum -> pval
    ssFy = np.cumsum(zFy,-2) /sumcorr[:,np.newaxis] # sum
    Ptgt_ssFy = softmax(ssFy)

    # pval -> combine
    # N.B. not numerically robust
    Ptgt_zFy = softmax(zFy)
    sPtgt_zFy = np.cumprod(Ptgt_zFy, -2) # combine
    sPtgt_zFy0 = sPtgt_zFy / np.sum(sPtgt_zFy, -1, keepdims=True) # normalize

    # alt computation,  softmax sum log(P)
    sPtgt_zFy =softmax(np.cumsum(np.log(Ptgt_zFy),-2))

    # incremental computation
    sPtgt_zFy = Ptgt_zFy.copy()
    for di in range(1,Ptgt_zFy.shape[-2]):
        #       prior            * likeihood
        postlikelihood = sPtgt_zFy[..., di-1, :] * sPtgt_zFy[..., di,:]
        #          post-likelihood / Z
        sPtgt_zFy[..., di,:] = postlikelihood / np.sum(postlikelihood,-1,keepdims=True)


    # N.B. correct way to do this is to integerate over the unit variance noise in the softmax? right?
    # Noise corrected computation, taking into account the error in Fy, which  scales as sqrt(N)
    # N.B. this is equivalent to having the scale in the softmax go as sqrt(N)
    if True:
        sPtgt_zFy =softmax(np.cumsum(np.log(Ptgt_zFy),-2)) # no summed error compenstation
    else:
        sPtgt_zFy =softmax(np.cumsum(np.log(Ptgt_zFy),-2)/sumcorr[:,np.newaxis]) # sum compensation

    plt.figure(3)
    plt.clf()
    p1=plt.subplot(311)
    plt.cla()
    plt.plot(ssFy[0,:,:])
    plt.grid(True)
    plt.title("summed scores")
    plt.subplot(312,sharex=p1)
    plt.cla()
    plt.plot(Ptgt_ssFy[0,:,:])
    plt.grid(True)
    plt.title("integerate->pVal")
    plt.subplot(313,sharex=p1)
    plt.cla()
    plt.title("pVal->integerate")
    plt.plot(sPtgt_zFy[0,:,:])
    plt.grid(True)
    plt.show()


def compute_pval_curve(X,pval):
    thresh=np.zeros((X.shape[1]))
    for t in range(X.shape[1]):
        Xt=X[:,t,:]
        for tst in np.linspace(np.min(Xt.ravel()),np.max(Xt.ravel()),40):
            if np.sum(Xt.ravel()>tst)/Xt.size < pval:
                thresh[t]=tst
                break
    return thresh

def compute_softmax_curve(X,pval,scale=3):
    thresh=np.ones((X.shape[1]))*np.nan
    for t in range(X.shape[1]):
        Xt=X[:,t,:]
        Z = np.sum(np.exp(scale*Xt),-1)
        for tst in np.linspace(np.min(Xt.ravel()),np.max(Xt.ravel())*2,40):
            softmax = np.exp(scale*tst) / (np.exp(scale*tst)+Z) # p-(tst==max)
            if np.mean(softmax.ravel()) > 1-pval:
                thresh[t]=tst
                break
    return thresh

def softmax_nout_corr(n):
    ''' approximate correction factor for probabilities out of soft-max to correct for number of outputs'''
    return min(2.45,1.25+np.log2(n)/5.5)

def softmax_vs_nout(nruns,nout,sscale):
    F = np.random.randn(nruns,np.max(nout)) # noise
    F0= np.random.randn(nruns,1) # signal noise
    Ftrue = np.linspace(0,5,50) # signal
    Ptrue = np.zeros((Ftrue.size,len(nout)))
    Pest  = np.zeros((Ftrue.size,len(nout)))
    for ni,nnoise in enumerate(nout):
        scale = sscale*softmax_nout_corr(nnoise)
        sexpF = np.sum(np.exp(scale*F[:,:nnoise]),-1) # sum exp-noise scores
        for ampi,sigamp in enumerate(Ftrue):
            # empherical estimate signal is biggest
            Ptrue[ampi,ni] = np.mean(np.all(sigamp+F0 > F[:,:nnoise],-1),0)
            # softmax estimate signal is biggest
            expF0 = np.exp( scale*(sigamp+F0) )
            softmax = expF0  /  (expF0 + sexpF[:,np.newaxis])
            # N.B. median so test the most-likely p-val est is similar to the true..
            Pest[ampi,ni] = np.median( softmax )
    # visualize the result....
    import matplotlib.pyplot as plt
    plt.clf();plt.subplot(221);
    plt.plot(Ptrue);plt.grid(True);plt.ylim([.8,1]);plt.xlabel('sig-amp');plt.subplot(222);
    plt.plot(Pest);plt.grid(True);plt.ylim([.8,1]);plt.xlabel('sig-amp'); plt.subplot(212);
    plt.plot(Ptrue-Pest);plt.legend(nout);plt.grid(True)
    return (Ptrue,Pest)

if __name__=='__main__':
    Fy=None
    # load plos-one dataset
    if True:
        from datasets import get_dataset
        from analyse_datasets import analyse_dataset
        l,f,_ = get_dataset('plos_one')
        X,Y,coords=l(f[0])
        _,_,Fy,_=analyse_dataset(X,Y,coords,evtlabs=('re','fe'),model='cca',rank=1)

    testcase(Fy)



import numpy as np
#@function
def zscore2Ptgt_softmax(f, softmaxscale=2, validTgt=None, marginalizemodels=True):
    '''
    convert normalized output scores into target probabilities
    Inputs:
     f - (nM,nTrl,nDecis,nY)  normalized accumulated scores]
     softmaxscale - float, slope to scale from scores to probabilities
     validtgtTrl = (nTrl,nY):bool which targets are valid in which trials 
    Outputs:
     Ptgt - (nTrl,nY) - target probability for each trial
    '''

    # fix the nuisance parameters to get per-output per trial score
    # pick the model

    # make 4d->3d for simplicity
    if f.ndim > 3:
        origf = f.copy()
        if f.shape[0] > 1:
            # WARNING: assume that have unique model for each output..
            # WARNING: f must have same mean and scale for this to be valid!!
            f=np.zeros(f.shape[1:])
            for mi in range(origf.shape[0]):
                f[..., mi]=origf[mi, :, :, mi]
        else:
            f = f[0, ...]
    elif f.ndim == 2:
        f = f[np.newaxis, :]
    elif f.ndim == 1:
        f = f[np.newaxis, np.newaxis, :]
    # Now : f= (nTrl,nDecis,nY)

    if validTgt is None: # get which outputs are used in which trials..
        validTgt = np.any(f != 0, 1) # (nTrl,nY)
    elif validTgt.ndim == 1:
        validTgt = validTgt[np.newaxis, :]

    noutcorr = softmax_nout_corr(np.sum(validTgt,1)) # (nTrl,)
    softmaxscale = softmaxscale * noutcorr[:,np.newaxis,np.newaxis] #(nTrl,1,1)

    # get the prob each output conditioned on the model
    Ptgteptimdl = softmax(f*softmaxscale,validTgt)
    
    if (marginalizemodels and
        (Ptgteptimdl.shape[1] > 1 or # mulitple decis points 
        (Ptgteptimdl.ndim > 3 and Ptgteptimdl.shape[0] > 1))): # multiple models
        # need to remove the nusiance parameters to get per-trial Y-prob
        ftgt=np.zeros((Ptgteptimdl.shape[0], Ptgteptimdl.shape[2])) # (nTrl,nY) [ nY x nTrl]
        for ti in range(Ptgteptimdl.shape[0]): # loop over trials
            Ptgtepmdl = Ptgteptimdl[ti, :, :] # (nDecis,nY) [ nY x nDecis ]
            # compute entropy over outputs for each decision point
            ent = 1 - np.sum(Ptgtepmdl*np.log(np.maximum(Ptgtepmdl, 1e-08)), 1) / -np.log(Ptgtepmdl.shape[1]) # (nDecis)
            # find when hit max-ent decison
            epidx = np.argmax(ent)
            wght = np.ones(Ptgtepmdl.shape[0]) # [ nDecis ]
            wght[:epidx] = 0
            # BODGE 1: just do weighted sum over decis points with more data than max-ent point
            ftgt[ti, :] = np.dot(wght, f[ti, :, :]) / np.sum(wght)

        Ptgt = np.exp(softmaxscale*(ftgt-np.max(ftgt, -1, keepdims=True))) # (nTrl,nY) [ nY x nTrl ]
        if not all(validTgt):
            Ptgt = Ptgt*validTgt
        Ptgt = Ptgt/np.sum(Ptgt, -1) # [nY x nTrl ]
    else:
        Ptgt = Ptgteptimdl

    if any(np.isnan(Ptgt.ravel())):
        if not all(np.isnan(Ptgt.ravel())):
            print('Error NaNs in target probabilities')
        Ptgt[:] = 0
    return Ptgt


def softmax(f,validTgt=None):
    ''' simple softmax over final dim of input array, with compensation for missing inputs with validTgt mask. '''
    Ptgteptimdl=np.exp(f-np.max(f, -1, keepdims=True)) # (nTrl,nDecis,nY) [ nY x nDecis x nTrl ]
    # cancel out the missing outputs
    if validTgt is not None and not all(validTgt.ravel()):
        Ptgteptimdl = Ptgteptimdl * validTgt[..., np.newaxis, :]
    # convert to softmax, with guard for div by zero
    Ptgteptimdl = Ptgteptimdl / np.maximum(np.sum(Ptgteptimdl, -1, keepdims=True),1e-6)
    return Ptgteptimdl

def softmax_nout_corr(n):
    ''' approximate correction factor for probabilities out of soft-max to correct for number of outputs'''
    #return np.minimum(2.45,1.25+np.log2(np.maximum(1,n))/5.5)/2.45
    return np.ones(n.shape) #np.minimum(2.45,1.25+np.log2(np.maximum(1,n))/5.5)/2.45


def calibrate_softmaxscale(f, validTgt=None, scales=(.5,1,1.5,2,2.5,3,3.5,4,5,7,10,15,20,30), MINP=.01):
    '''
    attempt to calibrate the scale for a softmax decoder to return calibrated probabilities
    Inputs:
     f - (nTrl,nDecis,nY) normalized accumulated scores]
     validTgt = (nTrl,nY) [nY x #Trl]:bool which targets are valid in which trials
     scales = set of possible soft-max scales to try
     MINP - float, minimium P-value.  We clip the true-target p-val to this level as a way
            of forcing the fit to concentrate on getting the p-val right when high, rather than
            over penalizing when it's wrong
    Outputs:
     softmaxscale - float, slope for softmax to return calibrated probabilities
    '''
    if validTgt is None: # get which outputs are used in which trials..
        validTgt = np.any(f != 0, 1) # (nTrl,nY)
    elif validTgt.ndim == 1:
        validTgt = validTgt[np.newaxis, :]
    # include the nout correction on a per-trial basis
    noutcorr = softmax_nout_corr(np.sum(validTgt,1)) # (nTrl,)

    # simply try a load of scales - as input are normalized shouldn't need more than this
    Ed = np.zeros(len(scales),)
    for i,s in enumerate(scales):
        softmaxscale = s * noutcorr[:,np.newaxis,np.newaxis] #(nTrl,1,1)

        # apply the soft-max with this scaling
        Ptgt = softmax(f*softmaxscale,validTgt)
        # Compute the loss = cross-entropy.  
        # As Y==0 is *always* the true-class, this becomes simply sum log this class 
        Ed[i] = np.sum(-np.log(np.maximum(Ptgt[...,0],MINP)))
        #print("{}) scale={} Ed={}".format(i,s,Ed[i]))
    # use the max-entropy scale
    mini = np.argmin(Ed)
    softmaxscale = scales[mini]
    print("softmaxscale={}".format(softmaxscale))
    return softmaxscale

#@function
def testcase():
    import numpy as np
    nY = 10
    nM = 1
    nEp = 340
    nTrl = 100
    noise = np.random.standard_normal((nM,nTrl,nEp,nY))
    noise = noise - np.mean(noise.ravel())
    noise = noise / np.std(noise.ravel())

    sigamp=0.25*np.ones(noise.shape[-2]) # [ nEp ]
    # no signal ast the start of the trial
  #startupNoise_samp=nEp*.5;
  #sigamp((1:size(sigamp,2))<startupNoise_samp)=0;
    Fy = np.copy(noise)
    # add the signal
    Fy[0, :, :, 0] = Fy[0, :, :, 0] + sigamp
    #print("Fy={}".format(Fy))
    
    sFy=np.cumsum(Fy,-2)
    from normalizeOutputScores import normalizeOutputScores
    ssFy,scale_sFy,N,_,_=normalizeOutputScores(Fy,minDecisLen=-1,filtLen=0)
    #print('ssFy={}'.format(ssFy.shape))
    from zscore2Ptgt_softmax import zscore2Ptgt_softmax, softmax
    smax = softmax(ssFy)
    #print("{}".format(smax.shape))
    Ptgt=zscore2Ptgt_softmax(ssFy,marginalizemodels=False) # (nTrl,nEp,nY)
    #print("Ptgt={}".format(Ptgt.shape))
    import matplotlib.pyplot as plt
    plt.clf()
    tri=1
    plt.subplot(411);plt.cla();
    plt.plot(sFy[0,tri,:,:])
    plt.plot(scale_sFy[0,tri,:],'k')
    plt.title('ssFy')
    plt.grid()
    plt.subplot(412);plt.cla();
    plt.plot(ssFy[0,tri,:,:])
    plt.title('ssFy')
    plt.grid()
    plt.subplot(413);plt.cla()
    plt.plot(smax[0,tri,:,:])
    plt.title('softmax')
    plt.grid()
    plt.subplot(414);plt.cla()
    plt.plot(Ptgt[tri,:,:])
    plt.title('Ptgt')
    plt.grid()
    plt.show()
    
    maxP=np.max(Ptgt,-1) # (nTrl,nEp) [ nEp x nTrl ]
    estopi=[ np.flatnonzero(maxP[tri,:]>.9)[-1] for tri in range(Ptgt.shape[0])]

if __name__=="__main__":
    testcase()



import os
import numpy as np
from scipy.io import loadmat, whosmat
from mindaffectBCI.decoder.utils import block_randomize, butter_sosfilt, upsample_codebook, lab2ind, window_axis

marker2stim=dict(lh=(1,3),rh=(2,4))

def load_twofinger(datadir, sessdir=None, sessfn=None, fs_out=60, stopband=((45,65),(0,1),(25,-1)), subtriallen=10, nvirt=20, verb=0, ch_idx=slice(32)):
    
    # load the data file
    Xfn = datadir
    if sessdir:
        Xfn = os.path.join(Xfn, sessdir)
    if sessfn:
        Xfn = os.path.join(Xfn, sessfn)
    sessdir = os.path.dirname(Xfn)

    data = loadmat(Xfn)

    def squeeze(v):
        while v.size == 1 and v.ndim > 0:
            v = v[0]
        return v

    fs = 512
    ch_names = [c[0] for c in squeeze(data['chann']).ravel()]
    X = squeeze(data['X'])  # (ch,samp)
    X = np.moveaxis(X,(0,1),(1,0)) # (samp,ch)
    X = X.astype(np.float32)
    X = np.ascontiguousarray(X)
    if ch_idx is not None:
        X = X[:, ch_idx]
        ch_names = ch_names[ch_idx]
    if verb>0: print("X={}".format(X.shape),flush=True)
    
    lab = squeeze(data['Y']).astype(int).ravel() # (samp,)

    # preprocess -> spectral filter, in continuous time!
    if stopband is not None:
        if verb > 0:
            print("preFilter: {}Hz".format(stopband))
        X, _, _ = butter_sosfilt(X,stopband,fs)
    
    # make the targets, for the events we care about
    Y, lab2class = lab2ind(lab,marker2stim.values()) # (nTrl, e) # feature dim per class
    if verb>0: print("Y={}".format(Y.shape))
        
    # preprocess -> downsample
    resamprate = int(fs/fs_out)
    if resamprate > 1:
        if verb > 0:
            print("resample: {}->{}hz rsrate={}".format(fs, fs/resamprate, resamprate))
        X = X[..., ::resamprate, :] # decimate X (trl, samp, d)
        # re-sample Y, being sure to keep any events in the re-sample window
        Y = window_axis(Y,winsz=resamprate,step=resamprate,axis=-2) # (trl, samp, win, e)
        Y = np.max(Y,axis=-2) # (trl,samp,e)  N.B. use max so don't loose single sample events
        fs = fs/resamprate
        if verb > 0:
            print("X={}".format(X.shape))
            print("Y={}".format(Y.shape))

    # make virtual targets
    Y = Y[:,np.newaxis,:] # (nsamp,1,e)
    Y_virt = block_randomize(Y, nvirt, axis=-3) # (nsamp,nvirt,e)
    Y = np.concatenate((Y, Y_virt), axis=-2) # (nsamp,1+nvirt,e)
    if verb>0: print("Y={}".format(Y.shape))

    # cut into sub-trials
    nsubtrials = X.shape[0]/fs/subtriallen
    if nsubtrials > 1:
        winsz = int(X.shape[0]//nsubtrials)
        if verb>0: print('subtrial winsz={}'.format(winsz))
        # slice into sub-trials
        X = window_axis(X,axis=0,winsz=winsz,step=winsz) # (trl,win,d)
        Y = window_axis(Y,axis=0,winsz=winsz,step=winsz) # (trl,win,nY)
        if verb>0: 
            print("X={}".format(X.shape))
            print("Y={}".format(Y.shape))
    
    # make coords array for the meta-info about the dimensions of X
    coords = [None]*X.ndim
    coords[0] = {'name':'trial'}
    coords[1] = {'name':'time','unit':'ms', \
                 'coords':np.arange(X.shape[1])/fs, \
                 'fs':fs}
    coords[2] = {'name':'channel','coords':ch_names}
    # return data + metadata
    return (X, Y, coords)

def testcase():
    import sys

    if os.path.isdir('D:\external_data'):
        sessfn = 'D:\\external_data\twente\twofinger\S00.mat'
    else:
        sessfn = '/home/jadref/data/bci/external_data/twente/twofinger/S00.mat'
    # command-line, for testing
    if len(sys.argv) > 1:
        sessfn = sys.argv[1]
 
    from load_twofinger import load_twofinger
    oX, oY, coords = load_twofinger(sessfn, fs_out=60, nsubtrials=40)
    times = coords[1]['coords']
    fs = coords[1]['fs']
    ch_names = coords[2]['coords']
    X=oX.copy()
    Y=oY.copy()

    print("X({}){}".format([c['name'] for c in coords],X.shape))
    print("Y={}".format(Y.shape))
    print("fs={}".format(fs))

    tau=fs*.7
    evtlabs = None
    times=np.arange(int(tau))/fs
    rank=1

    # visualize the dataset
    from stim2event import stim2event
    from updateSummaryStatistics import updateSummaryStatistics, plot_erp, plot_summary_statistics, idOutliers
    import matplotlib.pyplot as plt
    
    Cxx, Cxy, Cyy = updateSummaryStatistics(X, Y[...,0:1,:], tau=tau)

    plt.figure(1);
    print("summary stats")
    plot_summary_statistics(Cxx, Cxy, Cyy, evtlabs, times, ch_names)

    plt.figure(2);
    print("ERP")
    plot_erp(Cxy, ch_names=ch_names, evtlabs=evtlabs, times=times, plottype='plot', axis=-1)

    from model_fitting import MultiCCA
    from decodingCurveSupervised import decodingCurveSupervised
    cca = MultiCCA(tau=tau, evtlabs=evtlabs, rank=rank)
    scores = cca.cv_fit(X, Y)
    Fy = scores['estimator']
    print("Fy={}".format(Fy.shape))
    (_)=decodingCurveSupervised(Fy)

    # plot the solution
    from scoreStimulus import factored2full
    print("Plot Model")
    plt.figure(3)
    plot_erp(factored2full(cca.W_, cca.R_), ch_names=ch_names, evtlabs=evtlabs, times=times)
    #   plot Fy
    plt.figure(4)
    for ti in range(min(Fy.shape[0],25)):
        plt.subplot(5,5,ti+1)
        plt.imshow(np.cumsum(Fy[ti,:,:],axis=-2),aspect='auto')
    plt.show()
    
    
if __name__=="__main__":
    testcase()



import os
from glob import glob
import re
from mindaffectBCI.decoder.offline.load_mark_EMG import load_mark_EMG
from mindaffectBCI.decoder.offline.load_brainstream import load_brainstream
from mindaffectBCI.decoder.offline.load_p300_prn import load_p300_prn
from mindaffectBCI.decoder.offline.load_openBMI import load_openBMI
from mindaffectBCI.decoder.offline.load_cocktail import load_cocktail
from mindaffectBCI.decoder.offline.load_mTRF_audio import load_mTRF_audio
from mindaffectBCI.decoder.offline.load_twofinger import load_twofinger
from mindaffectBCI.decoder.offline.load_brainsonfire import load_brainsonfire
from mindaffectBCI.decoder.offline.load_ninapro_db2 import load_ninapro_db2
from mindaffectBCI.decoder.offline.load_mindaffectBCI import load_mindaffectBCI
from mindaffectBCI.decoder.utils import testSignal

def plos_one():
    '''generate the directory+filename info for the plos_one noisetagging dataset'''
    loader = load_brainstream # function to call to load the dataset
    if os.path.isdir(os.path.expanduser('~/data/bci')):
        dataroot = os.path.expanduser('~/data/bci/own_experiments/noisetagging_v3')
    else:
        dataroot = 'D:/own_experiments/noisetagging_v3/'
    sessdir = glob(os.path.join(dataroot, 's[0-9]*'))
    #sessdir = ['s{:d}'.format(i) for i in range(1, 12)]
    sessfn = 'traindata.mat'
    filenames = [os.path.join(dataroot, d, sessfn) for d in sessdir]
    return (loader, filenames, dataroot)

def lowlands():
    '''generate the directory+filename info for the lowlands noisetagging dataset'''
    loader = load_brainstream
    if os.path.isdir(os.path.expanduser('~/data/bci')):
    	dataroot = os.path.expanduser('~/data/bci/own_experiments/lowlands')
    else:
        dataroot = 'D:/own_experiments/lowlands/'
    filenames = glob(os.path.join(dataroot, '*_tr_train_1.mat'))
    return (loader, filenames, dataroot)

def p300_prn(label:str=None):
    '''generate dataset+filename for p300-prn2'''
    loader = load_p300_prn # function to call to load the dataset
    if os.path.isdir(os.path.expanduser('~/data/bci')):
        dataroot = os.path.expanduser('~/data/bci/')
    else:
        dataroot = 'D:/'
    expt = 'own_experiments/visual/p300_prn_2'
    filenames = glob(os.path.join(dataroot, expt, '*/*/jf_prep/*flash.mat')) + \
                glob(os.path.join(dataroot, expt, '*/*/jf_prep/*flip.mat'))
    if label is not None:
        filenames = [f for f in filenames if re.search(label,f)]
    return (loader, filenames, dataroot)

def tactileP3():
    '''generate dataset+filename for tactile P3'''
    loader = load_p300_prn # function to call to load the dataset
    if os.path.isdir(os.path.expanduser('~/data/bci')):
        dataroot = os.path.expanduser('~/data/bci/')
    else:
        dataroot = 'D:/'
    expt = 'own_experiments/tactile/P3speller'
    filenames = glob(os.path.join(dataroot, expt, '*/*/jf_prep/*offline.mat'))
    return (loader, filenames, dataroot)

def tactile_PatientStudy():
    '''generate dataset+filename for p300-prn2'''
    loader = load_p300_prn # function to call to load the dataset
    if os.path.isdir(os.path.expanduser('~/data/bci')):
        dataroot = os.path.expanduser('~/data/bci/')
    else:
        dataroot = 'D:/'
    expt = 'own_experiments/tactile/PatientStudy'
    filenames = glob(os.path.join(dataroot, expt, '*/*/jf_prep/*offline.mat'))
    return (loader, filenames, dataroot)

def openBMI(dstype="SSVEP"):
    loader = load_openBMI
    if os.path.isdir(os.path.expanduser('~/data/bci')):
        dataroot = os.path.expanduser('~/data/bci/external_data/gigadb/openBMI')
    else:
        dataroot= os.path.expanduser('D:/external_data/gigadb/openBMI')
    filenames = glob(os.path.join(dataroot, 'sess*/s*/sml_*'+ dstype + '.mat')) + \
                glob(os.path.join(dataroot, 'sml_*'+ dstype + '.mat'))
    return (loader, filenames, dataroot)

def twofinger():
    loader = load_twofinger
    if os.path.isdir(os.path.expanduser('~/data/bci')):
        dataroot = os.path.expanduser('~/data/bci/')
    else:
        dataroot= os.path.expanduser('D:/')
    exptdir = 'external_data/twente/twofinger'
    filenames =  glob(os.path.join(dataroot, exptdir, 'S??.mat'))
    return (loader, filenames, dataroot)

def brains_on_fire_online():
    loader = load_brainsonfire
    if os.path.isdir(os.path.expanduser('~/data/bci')):
        dataroot = os.path.expanduser('~/data/bci/')
    else:
        dataroot= os.path.expanduser('D:/')
    exptdir = 'own_experiments/motor_imagery/brainsonfire/brains_on_fire_online'
    filenames =  glob(os.path.join(dataroot, exptdir, 'subject*/raw_buffer/0001'))
    return (loader, filenames, dataroot)


def mTRF_audio():
    loader = load_mTRF_audio
    if os.path.isdir(os.path.expanduser('~/data/bci')):
        dataroot= os.path.expanduser('~/data/bci/external_data/mTRF')
    else:
        dataroot= os.path.expanduser('D:/external_data/mTRF')
    filenames = [os.path.join(dataroot, 'speech_data.mat')]
    return (loader, filenames, dataroot)

def ninapro_db2():
    loader = load_ninapro_db2
    if os.path.isdir(os.path.expanduser("~/data")):
        dataroot = os.path.expanduser('~/data/bci/')
    else:
        dataroot = "D:/"
    exptdir="external_data/ninapro"
    filenames = glob(os.path.join(dataroot, exptdir, 's*', '*E1*.mat'))
    return (loader, filenames, dataroot)

def cocktail():
    loader = load_cocktail
    if os.path.isdir(os.path.expanduser('~/data/bci')):
        dataroot= os.path.expanduser('~/data/bci/')
    else:
        dataroot= os.path.expanduser('D:/')
    exptdir="external_data/dryad/Cocktail Party"
    filenames = glob(os.path.join(dataroot, exptdir, 'EEG', 'Subject*'))
    return (loader,filenames,dataroot)

def mark_EMG():
    loader = load_mark_EMG
    if os.path.isdir(os.path.expanduser('~/data/bci')):
        dataroot= os.path.expanduser('~/data/bci/')
    else:
        dataroot= os.path.expanduser('D:/')
    exptdir="own_experiments/emg/facial"
    filenames = glob(os.path.join(dataroot, exptdir, 'training_data_SV_*.mat'))
    return (loader,filenames,dataroot)    

def testdataset(fn, **args):
    '''  a conforming toy dataset loader'''
    fs=10
    X,Y,st,A,B=testSignal(**args)
    # make coords array for the meta-info about the dimensions of X
    coords = [None]*X.ndim
    coords[0] = {'name':'trial'}
    coords[1] = {'name':'time','unit':'ms', \
                 'coords':[i*1000/fs for i in range(X.shape[1])], \
                 'fs':fs}
    coords[2] = {'name':'channel','coords':None}
    return (X, Y, coords)

def mindaffectBCI(exptdir, **args):
    loader = load_mindaffectBCI
    filenames = glob(os.path.join(exptdir, 'mindaffectBCI*.txt'))
    return (loader,filenames,exptdir)

def toy():
    ''' make a toy dataset for testing '''
    loader = testdataset
    filenames = [None]
    dataroot = None
    return (loader,filenames,dataroot)

def dataset_generator(dataset, **kwargs):
    ''' generator for individual datasets from the given set of datasets '''
    loadfn,  filenames, dataroot = get_dataset(dataset)
    for fn in filenames:
        X, Y, coords = loadfn(fn, **kwargs)
        yield (X, Y, coords, fn)

def test_loader(loadfn, filenames, dataroot, **kwargs):
    for fn in filenames:
        try:
            X, Y, coords= loadfn(fn, **kwargs)
            print("ds={}\nX({})={}\nY={}\n".format(fn, [c['name'] for c in coords], X.shape, Y.shape))
        except Exception as ex:
            print("ds={}\nFAILED\n{}".format(fn, ex))

def get_dataset(dsname,*args, **kwargs):
    if dsname == 'openBMI_SSVEP':
        return openBMI("SSVEP")
    elif dsname == 'openBMI_ERP':
        return openBMI("ERP")
    elif dsname == 'openBMI_MI':
        return openBMI("MI")
    else:
        try:
            return eval(dsname)(*args,**kwargs)
        except:
            raise NotImplementedError("don't know dataset {}".format(dsname))

def testcase():
    datasets=["openBMI_MI","tactileP3","toy","mark_EMG","brainsonfire","twofinger","ninapro_db2","openBMI_MI","openBMI_ERP","openBMI_SSVEP","cocktail","lowlands","plos_one",'p300_prn',"mTRF_audio"]
    for d in datasets:
        print(d)
        #try:
        loadfn, filenames,  dataroot = get_dataset(d)
        test_loader(loadfn, filenames, dataroot)
        #except:
        #    print("Error with dataset {}".format(d))

if __name__ == '__main__':
    testcase()



#!/usr/bin/env python3
#  Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
from mindaffectBCI.utopiaclient import UtopiaClient,Subscribe,DataPacket,SignalQuality
from collections import deque

# rate at which to update the display
updatet_ms=50
# amout of data to display
datawindow_ms=5000
# gap in data units between channel lines
linespace_uv=20

def makeLines(ax, nch):
  line = [None]*nch
  bbox = [None]*nch
  h = ax.get_ylim()[1]/nch
  x = ax.get_xlim()[0]
  w = ax.get_xlim()[1]-x
  for li in range(nch):
    y = li*h
    bbox[li] = (x, y, w, h)
    line[li], = ax.plot(0, 0)
  return (line, bbox)

# init the display
fig=plt.figure(1)
fig.clear()
ax=fig.add_subplot(111) # single fullfig axes

ax.set_xlim(-datawindow_ms, 0) # xlim set to match data duration
ax.set_xlabel('time (ms)')
ax.set_ylabel('uV')

# make lines and bounding boxes for the lines
line=[]
bbox=[]
fig.show()

# connect to the mindaffectDecoder
U=UtopiaClient()
U.autoconnect(timeout_ms=5000, queryifhostnotfound=True)
# subscribe to the raw-data messages
U.sendMessage(Subscribe(None, "D"))

# render loop
qual=[]
dataringbuffer = deque()
t0 = U.getTimeStamp()
while True:
  # get enough new data for update window
  t=U.getTimeStamp()
  while U.getTimeStamp() < t+updatet_ms: 
    msgs = U.getNewMessages(0)
    for m in msgs:
      if m.msgID == DataPacket.msgID:
        print('D', end='', flush=True)
        if U.getTimeStamp() > t0+datawindow_ms: # slide buffer
          dataringbuffer.popleft()
        dataringbuffer.append(m.samples)
      elif m.msgID == SignalQuality.msgID:
        qual = m.qual
  # draw the lines
  if dataringbuffer:
    nch=len(dataringbuffer[0][0])
    if not len(line) == nch: # update the line set if number channel changed
      ax.set_ylim(-1*linespace_uv,nch*linespace_uv) # ylim are set to match uV
      line, bbox = makeLines(ax,nch)
    for ci in range(nch):
      # flatten into list samples for each channel
      d = [t[ci] for m in dataringbuffer for t in m]
      # map to screen coordinates
      yscale = bbox[ci][3]/linespace_uv # 10 uV between lines
      mu = sum(d)/len(d) # center
      y = [bbox[ci][1] + (s-mu)*yscale for s in d]
      x = [bbox[ci][0] + bbox[ci][2]*i/len(y) for i in range(len(y))]
      line[ci].set_data(x,y)
      # add qual info
      
  # force plot to re-draw 
  fig.canvas.draw()
  fig.canvas.flush_events()



#!/usr/bin/env python3

#  Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Import the utopia2output module
from mindaffectBCI.utopia2output import Utopia2Output
u2o=Utopia2Output()
u2o.connect()
def helloworld(objID):
   print("hello world")
# set the objectID2Action dictionary to use our helloworld function if 1 is selected 
u2o.objectID2Action={ 1:helloworld }
u2o.run()



#!/usr/bin/env python3

from utopia2output import *
from time import sleep
from pyb00st.movehub import MoveHub
from pyb00st.constants import *

mymovehub = MoveHub('00:16:53:A5:03:67', 'Auto', 'hci0')
mymovehub.start()
while True:  
    try:
        mymovehub.connect()
    except Exception as e:
        print(e)
        print('Cannot connect to Boost, trying again..')
    else:
        break

# define our action functions
def head (objID):
    print('move head')
    try:
        mymovehub.run_motor_for_angle(MOTOR_D, 70, 20)
        sleep(2)
        mymovehub.run_motor_for_angle(MOTOR_D, 140, -10)
        sleep(1)
        mymovehub.run_motor_for_angle(MOTOR_D, 70, 20)
    except:
        print('Cannot connect to Boost.')

def left (objID):
    print('left')
    try:
        mymovehub.run_motors_for_angle(MOTOR_AB, 180, -10, 10)
        sleep(3)
        mymovehub.run_motors_for_angle(MOTOR_AB, 180, 10, -10)
        sleep(1)
    except:
        print('Cannot connect to Boost.')
def right (objID):
    print('right')
    try:
        mymovehub.run_motors_for_angle(MOTOR_AB, 180, 10, -10)
        sleep(3)
        mymovehub.run_motors_for_angle(MOTOR_AB, 180, -10, 10)
        sleep(1)
    except:
        print('Cannot connect to Boost.')


# IMPORTANT: Somewhere:
# mymovehub.stop()
#We then connect them to their trigger object identifiers
objectID2Action = { 52 :head, 50 :left, 51 :right}
# create the output object
utopia2output=Utopia2Output(outputPressThreshold=None)
# replace the default action dictionary with ours. N.B. be careful
utopia2output.objectID2Action=objectID2Action
# run the output module
utopia2output.run()



#!/usr/bin/env python3
#
#  Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import struct
import time
import socket
import sys

class UtopiaMessage:
    """Class for a generic UtopiaMessage, i.e. the common structure of all messages """
    def __init__(self, msgID=0, msgName=None, version=0):
        self.msgID   = msgID
        self.msgName = msgName
        self.version = version

    def __str__(self):
        return '%i\n'%(self.msgID)

class TimeStampClock:
    """ Base class for time-stamp sources.  Match this prototype to replace the default timestamp source """
    def getTimeStamp(self):
        ''' get the time-stamp in milliseconds ! N.B. **must** fit in an int32! '''
        return (int(time.perf_counter()*1000) % (1<<31))

class RawMessage(UtopiaMessage):
    msgName="RAW"
    """Class for a raw utopia message, i.e. decoded header but raw payload"""
    def __init__(self, msgID, version, payload):
        """Construct a raw message from given msgID and payload"""
        super().__init__(msgID, RawMessage.msgName, version)
        self.payload = payload
        if hasattr(self.payload, 'encode'):
            self.payload = self.payload.encode()
        if not isinstance(self.payload, bytes):
            raise Exception("Illegal-value")
        
    @classmethod
    def fromUtopiaMessage(cls, msg):
        """Construct a raw-message wrapper from a normal payload message"""
        return cls(msg.msgID, msg.version, msg.serialize())

    def __str__(self):
        return '%c(%d) [%i]\n'%(chr(self.msgID), self.msgID, len(self.payload))
    
    def serialize(self):
        """convert the raw message to the string to make it network ready"""
        S= struct.pack("<BBH", self.msgID, self.version, len(self.payload))
        S = S + self.payload
        return S

    @classmethod
    def deserialize(cls, buf):
        """Read a raw message from a byte-buffer, return the read message and the number of bytes used from the bytebuffer, or None, 0 if message is mal-formed"""
        bufsize = len(buf)
        if bufsize < 4:
            print("Buffer too short for header")
            return (None, 0)
        (msgID, ver, msgsize) = struct.unpack('<BBH', buf[0:4])
        # read the rest of the message
        if msgsize > 0:
            if bufsize >= 4+msgsize:
                payload = buf[4:4+msgsize]
            else:
                print("Buffer too short for payload: id:{}, ver:{}, sz:{}".format(chr(msgID), ver, msgsize))
                return (None, 0)
        else:
            payload = None
        msg=cls(msgID, ver, payload)
        return (msg, 4+msgsize)        

    @classmethod
    def deserializeMany(cls, buf):
        """decode multiple RawMessages from the byte-buffer of data, return the length of data consumed."""
        msgs=[]
        nconsumed=0
        while nconsumed < len(buf):            
            (msg, msgconsumed)=RawMessage.deserialize(buf[nconsumed:])
            if  msg==None or msgconsumed == 0:
                break # bug-out if invalid/incomplete message
            msgs.append(msg)
            nconsumed = nconsumed + msgconsumed
        return (msgs, nconsumed)
    
    
class Heartbeat(UtopiaMessage):    
    """ the HEARTBEAT utopia message class """

    # Static definitions of the class type constants
    msgID=ord('H')
    msgName="HEARTBEAT"
    
    def __init__(self, timestamp=None, statemessage:str=None):
        super().__init__(Heartbeat.msgID, Heartbeat.msgName)
        self.timestamp=timestamp
        self.statemessage = statemessage
        if self.statemessage is not None: # v1 heartbeat with state message string
            self.version = 1

    def serialize(self):
        """Returns the contents of this event as a string, ready to send over the network, 
           or None in case of conversion problems.
        """
        S = struct.pack('<i', int(self.timestamp))
        if self.statemessage is not None:
            S = S + bytes(self.statemessage, 'utf8')
        return S

    @staticmethod
    def deserialize(buf):
        """Static method to create a HEARTBEAT class from a **PAYLOAD** byte-stream, return the number of bytes consumed from buf"""
        bufsize = len(buf)
        if bufsize < 4:
            return (None, 0)
        (timestamp, ) = struct.unpack('<i', buf[0:4])
        if bufsize>4:
            statemessage = buf[4:].decode('utf-8')
            return (Heartbeat(timestamp,statemessage), len(buf))
        else:
            return (Heartbeat(timestamp), 4)
    
    def __str__(self):
        return "%c(%d) %s %i"%(self.msgID, self.msgID, self.msgName, self.timestamp)


class StimulusEvent(UtopiaMessage):
    """ the STIMULUEVENT utopia message class """
    
    # Static definitions of the class type constants
    msgID=ord('E')
    msgName="STIMULUSEVENT"

    def __init__(self, timestamp=None, objIDs=None, objState=None):
        super().__init__(StimulusEvent.msgID, StimulusEvent.msgName)
        self.timestamp=timestamp
        self.objIDs=objIDs
        self.objState=objState

    def serialize(self):
        """Converts this message to a string representation to send over the network
        """
        S = struct.pack("<i", int(self.timestamp)) # timestamp
        S = S + struct.pack("<B", len(self.objIDs))  # nObj
        for objid, objstate in zip(self.objIDs, self.objState):
            S = S + struct.pack("<BB", int(objid), int(objstate)) # [objID, objState] pairs
        return S

    @staticmethod
    def deserialize(buf):
        """Static method to create a STIMULUSEVENT class from a **PAYLOAD** byte-stream, return created object and the number of bytes consumed from buf"""
        bufsize = len(buf)
        if bufsize < 5:
            return (None, 0)
        (timestamp, nobj) = struct.unpack("<iB", buf[0:5])
        if bufsize < 5+nobj*2:
            return (None, 0)
        objIDs=[]
        objState=[]
        for i in range(5, len(buf), 2):
            (objid, objstate)=struct.unpack("<BB", buf[i:i+2])
            objIDs.append(objid)
            objState.append(objstate)
        msg=StimulusEvent(timestamp, objIDs, objState)
        return (msg, 5+nobj*2)

    def __str__(self):
        return "%s %i"%(self.msgName, self.timestamp) + "".join("(%i, %i)"%(x, y) for x, y in zip(self.objIDs, self.objState))

class PredictedTargetProb(UtopiaMessage):    
    """ the PREDICTEDTARGETPROB utopia message class """

    # Static definitions of the class type constants
    msgID=ord('P')
    msgName="PREDICTEDTARGETPROB"
    
    def __init__(self, timestamp=None, Yest=0, Perr=1.0):
        super().__init__(PredictedTargetProb.msgID, PredictedTargetProb.msgName)
        self.timestamp=timestamp
        self.Yest     =Yest
        self.Perr     =Perr

    def serialize(self):
        """Returns the contents of this event as a string, ready to send over the network, 
           or None in case of conversion problems.
        """
        S = struct.pack('<ibf', int(self.timestamp), int(self.Yest), self.Perr)
        return S

    @staticmethod
    def deserialize(buf):
        """Static method to create a MODECHANGE class from a **PAYLOAD** byte-stream, return the number of bytes consumed from buf"""
        bufsize = len(buf)
        if bufsize < 4:
            return (None, 0)
        (timestamp, Yest, Perr) = struct.unpack('<ibf', buf[:(4+1+4)])
        msg = PredictedTargetProb(timestamp, Yest, Perr)
        return (msg, 4)
    
    def __str__(self):
        return "%c(%d) %s %i Yest=%d Perr=%f"%(self.msgID, self.msgID, self.msgName, self.timestamp, self.Yest, self.Perr)



class PredictedTargetDist(UtopiaMessage):
    """ the PredictedTargetDist utopia message class """
    
    # Static definitions of the class type constants
    msgID=ord('F')
    msgName="PREDICTEDTARGETDIST"

    def __init__(self, timestamp=None, objIDs=None, pTgt=None):
        super().__init__(PredictedTargetDist.msgID, PredictedTargetDist.msgName)
        self.timestamp=timestamp
        self.objIDs=objIDs
        self.pTgt=pTgt

    def serialize(self):
        """Converts this message to a string representation to send over the network
        """
        S = struct.pack("<i", int(self.timestamp)) # timestamp
        S = S + struct.pack("<B", len(self.objIDs))  # nObj
        for objid, pTgt in zip(self.objIDs, self.pTgt):
            S = S + struct.pack("<Bf", int(objid), float(pTgt)) # [objID, pTgt] pairs
        return S

    @staticmethod
    def deserialize(buf):
        """Static method to create a PREDICTEDTARGETDIST class from a **PAYLOAD** byte-stream, return created object and the number of bytes consumed from buf"""
        bufsize = len(buf)
        if bufsize < 5:
            return (None, 0)
        (timestamp, nobj) = struct.unpack("<iB", buf[0:5])
        if bufsize < 5+nobj*(1+4):
            return (None, 0)
        objIDs=[]
        pTgt=[]
        for i in range(5, len(buf), (1+4)):
            (objid, ptgt)=struct.unpack("<Bf", buf[i:i+(1+4)])
            objIDs.append(objid)
            pTgt.append(ptgt)
        msg=PredictedTargetDist(timestamp, objIDs, pTgt)
        return (msg, 5+nobj*2)

    def __str__(self):
        return "%s %i"%(self.msgName, self.timestamp) + "".join("(%i, %f)"%(x, y) for x, y in zip(self.objIDs, self.pTgt))


class DataPacket(UtopiaMessage):
    """ the DATAPACKET utopia message class """
    
    # Static definitions of the class type constants
    msgID=ord('D')
    msgName="DATAPACKET"

    def __init__(self, timestamp=None, samples=None):
        super().__init__(DataPacket.msgID, DataPacket.msgName)
        self.timestamp=timestamp
        self.samples=samples

    def serialize(self):
        """Converts this message to a string representation to send over the network
        """
        S = struct.pack("<i", int(self.timestamp)) # timestamp
        S = S + struct.pack("<i", len(self.samples))  # nsamp
        for tp in self.samples:
            S = S + struct.pack("<%df"%(len(tp)), *tp)
        return S

    @staticmethod
    def deserialize(buf):
        """Static method to create a STIMULUSEVENT class from a **PAYLOAD** byte-stream, return created object and the number of bytes consumed from buf"""
        bufsize = len(buf)
        if bufsize < 8:
            return (None, 0)
        timestamp, nsamp = struct.unpack("<ii", buf[0:8])
        nch=int((bufsize-8)/(nsamp*4))
        samples=[]
        for t in range(nsamp):
            sampt=struct.unpack_from("<%df"%(nch), buf, 8+t*4*nch)
            samples.append(sampt)
        msg=DataPacket(timestamp, samples)
        return (msg, 8+nsamp*nch*8)

    def __str__(self):
        ss="%c(%d) %s %i "%(self.msgID, self.msgID, self.msgName, self.timestamp)
        ss= ss+ "[%dx%d]"%(len(self.samples), len(self.samples[0]))
        for chs in self.samples:
            chstr = "".join(["%f, "%(c) for c in chs])
            ss = ss + "["+chstr+"]"
        return  ss


class DataHeader(UtopiaMessage):
    """ the DATAHEADER utopia message class """
    
    # Static definitions of the class type constants
    msgID=ord('A')
    msgName="DATAHEADER"

    def __init__(self, timestamp=None, fsample=None, nchannels=None, labels=None):
        super().__init__(DataHeader.msgID, DataHeader.msgName)
        self.timestamp=timestamp
        self.fsample=fsample
        self.nchannels=nchannels
        self.labels=labels

    def serialize(self):
        """Converts this message to a string representation to send over the network
        """
        S = struct.pack("<i", int(self.timestamp)) # timestamp
        S = S + struct.pack("<i", self.nchannels)
        S = S + struct.pack("<f", self.fsample)
        if self.labels: 
            # comma separated list of channel names
            S = S + bytes(", ".join(self.labels), 'utf-8')
        return S

    @staticmethod
    def deserialize(buf):
        """Static method to create a HEADER class from a **PAYLOAD** byte-stream, return created object and the number of bytes consumed from buf"""
        bufsize = len(buf)
        if bufsize < 12:
            return (None, 0)
        timestamp, nchannels, fsample = struct.unpack("<iif", buf[0:8])
        labels = buf[4:].decode('utf-8')
        labels = labels.split(", ")
        msg=DataHeader(timestamp, nchannels, fsample, labels)
        return (msg, bufsize)

    def __str__(self):
        ss="%c(%d) %s %i "%(self.msgID, self.msgID, self.msgName, self.timestamp)
        ss= ss+ "%dch @ %gHz"%(self.nchannels, self.fsample)
        ss= ss+ ", ".join(self.labels)
        return  ss

class NewTarget(UtopiaMessage):    
    """ the NEWTARGET utopia message class """

    # Static definitions of the class type constants
    msgID=ord('N')
    msgName="NEWTARGET"
    
    def __init__(self, timestamp=None):
        super().__init__(NewTarget.msgID, NewTarget.msgName)
        self.timestamp=timestamp

    def serialize(self):
        """Returns the contents of this event as a string, ready to send over the network, 
           or None in case of conversion problems.
        """
        S = struct.pack('<i', int(self.timestamp))
        return S

    @staticmethod
    def deserialize(buf):
        """Static method to create a HEARTBEAT class from a **PAYLOAD** byte-stream, return the number of bytes consumed from buf"""
        bufsize = len(buf)
        if bufsize < 4:
            return (None, 0)
        (timestamp, ) = struct.unpack('<i', buf[0:4])
        msg = NewTarget(timestamp)
        return (msg, 4)
    
    def __str__(self):
        return "%c(%d) %s %i"%(self.msgID, self.msgID, self.msgName, self.timestamp)

class Selection(UtopiaMessage):    
    """ the SELECTION utopia message class """

    # Static definitions of the class type constants
    msgID=ord('S')
    msgName="SELECTION"
    
    def __init__(self, timestamp=None, objID=None):
        super().__init__(Selection.msgID, Selection.msgName)
        self.timestamp=timestamp
        self.objID    =objID

    def serialize(self):
        """Returns the contents of this event as a string, ready to send over the network, 
           or None in case of conversion problems.
        """
        S = struct.pack('<i', int(self.timestamp))
        S = S + struct.pack('<B', self.objID)
        return S

    @staticmethod
    def deserialize(buf):
        """Static method to create a SELECTION class from a **PAYLOAD** byte-stream, return the number of bytes consumed from buf"""
        bufsize = len(buf)
        if bufsize < 5:
            return (None, 0)
        (timestamp, objID) = struct.unpack('<iB', buf[0:5])
        msg = Selection(timestamp, objID)
        return (msg, 5)
    
    def __str__(self):
        return "%c(%d) %s %i id:%i"%(self.msgID, self.msgID, self.msgName, self.timestamp, self.objID)
    
class Reset(UtopiaMessage):    
    """ the RESET utopia message class """

    # Static definitions of the class type constants
    msgID=ord('R')
    msgName="RESET"
    
    def __init__(self, timestamp=None):
        super().__init__(Reset.msgID, Reset.msgName)
        self.timestamp=timestamp

    def serialize(self):
        """Returns the contents of this event as a string, ready to send over the network, 
           or None in case of conversion problems.
        """
        S = struct.pack('<i', int(self.timestamp))
        return S

    @staticmethod
    def deserialize(buf):
        """Static method to create a HEARTBEAT class from a **PAYLOAD** byte-stream, return the number of bytes consumed from buf"""
        bufsize = len(buf)
        if bufsize < 4:
            return (None, 0)
        (timestamp, ) = struct.unpack('<i', buf[0:4])
        msg = Reset(timestamp)
        return (msg, 4)
    
    def __str__(self):
        return "%c(%d) %s %i"%(self.msgID, self.msgID, self.msgName, self.timestamp)

    

class ModeChange(UtopiaMessage):    
    """ the MODECHANGE utopia message class """

    # Static definitions of the class type constants
    msgID=ord('M')
    msgName="MODECHANGE"
    
    def __init__(self, timestamp=None, newmode=None):
        super().__init__(ModeChange.msgID, ModeChange.msgName)
        self.timestamp=timestamp
        self.newmode  =newmode

    def serialize(self):
        """Returns the contents of this event as a string, ready to send over the network, 
           or None in case of conversion problems.
        """
        S = struct.pack('<i', int(self.timestamp))
        S = S + bytes(self.newmode, 'utf-8')
        return S

    @staticmethod
    def deserialize(buf):
        """Static method to create a MODECHANGE class from a **PAYLOAD** byte-stream, return the number of bytes consumed from buf"""
        bufsize = len(buf)
        if bufsize < 4:
            return (None, 0)
        (timestamp, ) = struct.unpack('<i', buf[0:4])
        newmode = buf[4:].decode('utf-8')
        msg = ModeChange(timestamp, newmode)
        return (msg, 4)
    
    def __str__(self):
        return "%c(%d) %s %i %s"%(self.msgID, self.msgID, self.msgName, self.timestamp, self.newmode)

class Log(UtopiaMessage):    
    """ the LOG utopia message class """

    # Static definitions of the class type constants
    msgID=ord('L')
    msgName="LOG"
    
    def __init__(self, timestamp=None, logmsg=None):
        super().__init__(Log.msgID, Log.msgName)
        self.timestamp=timestamp
        self.logmsg  =logmsg

    def serialize(self):
        """Returns the contents of this event as a string, ready to send over the network, 
           or None in case of conversion problems.
        """
        S = struct.pack('<i', int(self.timestamp))
        S = S + bytes(self.logmsg, 'utf-8')
        return S

    @staticmethod
    def deserialize(buf):
        """Static method to create a MODECHANGE class from a **PAYLOAD** byte-stream, return the number of bytes consumed from buf"""
        bufsize = len(buf)
        if bufsize < 4:
            return (None, 0)
        (timestamp, ) = struct.unpack('<i', buf[0:4])
        logmsg = buf[4:].decode('utf-8')
        msg = ModeChange(timestamp, logmsg)
        return (msg, 4)
    
    def __str__(self):
        return "%c(%d) %s %i %s"%(self.msgID, self.msgID, self.msgName, self.timestamp, self.logmsg)


class SignalQuality(UtopiaMessage):    
    """ the SIGNALQUALITY utopia message class """

    # Static definitions of the class type constants
    msgID=ord('Q')
    msgName="SIGNALQUALITY"
    
    def __init__(self, timestamp=None, signalQuality=None):
        super().__init__(SignalQuality.msgID, SignalQuality.msgName)
        self.timestamp=timestamp
        self.signalQuality=signalQuality

    def serialize(self):
        """Returns the contents of this event as a string, ready to send over the network, 
           or None in case of conversion problems.
        """
        S = struct.pack('<i', int(self.timestamp))
        S = S + b''.join([ struct.pack('<f', float(q)) for q in self.signalQuality ])
        return S

    @staticmethod
    def deserialize(buf):
        """Static method to create a SIGNALQUALIYT class from a **PAYLOAD** byte-stream, return the number of bytes consumed from buf"""
        bufsize = len(buf)
        if bufsize < 4:
            return (None, 0)
        (timestamp, ) = struct.unpack('<i', buf[0:4])
        signalQuality = struct.unpack_from('<%df'%((bufsize-4)//4), buf, 4)
        #                  , buf[i:i+4])[0] for i in range(4, bufsize, 4) ]
        msg = SignalQuality(timestamp, signalQuality)
        return (msg, bufsize)
    
    def __str__(self):
        return "%c(%d) %s %i [%s]"%(self.msgID, self.msgID, self.msgName, self.timestamp, ", ".join(["%f"%q for q in self.signalQuality]))

class Subscribe(UtopiaMessage):    
    """ the SUBSCRIBE utopia message class """

    # Static definitions of the class type constants
    msgID=ord('B')
    msgName="SUBSCRIBE"
    
    def __init__(self, timestamp=None, messageIDs=None):
        super().__init__(Subscribe.msgID, Subscribe.msgName)
        self.timestamp=timestamp
        self.messageIDs=messageIDs

    def serialize(self):
        """Returns the contents of this event as a string, ready to send over the network, 
           or None in case of conversion problems.
        """
        S = struct.pack('<i', int(self.timestamp))
        if type(self.messageIDs) is str:#str
            S = S + bytes(self.messageIDs, 'utf-8')
        else:# int->byte
            S = S + b''.join([ struct.pack('<b', byte(q)) for q in self.messageIDs ])
        return S

    @staticmethod
    def deserialize(buf):
        """Static method to create a SIGNALQUALITY class from a **PAYLOAD** byte-stream, return the number of bytes consumed from buf"""
        bufsize = len(buf)
        if bufsize < 4:
            return (None, 0)
        (timestamp, ) = struct.unpack('<i', buf[0:4])
        #messageIDs = [ struct.unpack('<i', buf[i:i+4]) for i in range(4, bufsize, 4) ]
        messageIDs = struct.unpack_from('<%di'%((bufsize-4)//4), buf, 4)
        msg = Subscribe(timestamp, messageIDs)
        return (msg, bufsize)
    
    def __str__(self):
        return "%c(%d) %s %i [%s]"%(self.msgID, self.msgID, self.msgName, self.timestamp, "{}".format(self.messageIDs))

    
# Helper functions for dealing with raw-messages -> classes    
def decodeRawMessage(msg):
    """decode utopia RawMessages into actual message objects"""
    if msg.msgID==StimulusEvent.msgID:
        decodedmsg, _= StimulusEvent.deserialize(msg.payload)
    elif msg.msgID==Heartbeat.msgID:
        decodedmsg, _ = Heartbeat.deserialize(msg.payload)
    elif msg.msgID==PredictedTargetProb.msgID:
        decodedmsg, _ = PredictedTargetProb.deserialize(msg.payload)
    elif msg.msgID==PredictedTargetDist.msgID:
        decodedmsg, _ = PredictedTargetDist.deserialize(msg.payload)
    elif msg.msgID==Selection.msgID:
        decodedmsg, _ = Selection.deserialize(msg.payload)
    elif msg.msgID==ModeChange.msgID:
        decodedmsg, _ = ModeChange.deserialize(msg.payload)
    elif msg.msgID==NewTarget.msgID:
        decodedmsg, _ = NewTarget.deserialize(msg.payload)
    elif msg.msgID==SignalQuality.msgID:
        decodedmsg, _ = SignalQuality.deserialize(msg.payload)
    elif msg.msgID==Reset.msgID:
        decodedmsg, _ = Reset.deserialize(msg.payload)
    elif msg.msgID==DataPacket.msgID:
        decodedmsg, _ = DataPacket.deserialize(msg.payload)
    elif msg.msgID==Subscribe.msgID:
        decodedmsg, _ = Subscribe.deserialize(msg.payload)
    else:
        decodedmsg = msg
    return decodedmsg

def decodeRawMessages(msgs):
    """decode utopia RawMessages into actual message objects"""
    return [ decodeRawMessage(msg) for msg in msgs ]

def ssdpDiscover(servicetype=None, timeout=3, numretries=1):
    '''auto-discover the utopia-hub using ssdp discover messages'''
    ssdpgroup = ("239.255.255.250", 1900)
    msearchTemplate = "\r\n".join([
        'M-SEARCH * HTTP/1.1', 
        'HOST: {0}:{1}', 
        'MAN: "ssdp:discover"', 
        'ST: {st}', 'MX: {mx}', '', ''])
    # make and send the discovery message
    service= servicetype if servicetype is not None else "ssdp:all"
    msearchMessage = msearchTemplate.format(*ssdpgroup, st=service, mx=timeout)
    if sys.version_info[0] == 3: msearchMessage = msearchMessage.encode("utf-8")
    try:
        # make the UDP socket to the multicast group with timeout
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_LOOP, 1)
        sock.settimeout(timeout)
    except: # abort if cant setup socket
        return () 
    responses=[]
    for _ in range(numretries):
        print("Sending query message:\n%s"%(msearchMessage))
        try:
            sock.sendto(msearchMessage, ssdpgroup)
        except: # abort if send fails
            return ()
        # wait for responses and store the location info for the matching ones
        print("Waiting responses")
        try:
            rsp, addr=sock.recvfrom(8192)
            rsp=rsp.decode('utf-8')
            print("Got response from: %s\n%s\n"%(addr, rsp))
            # does the response contain servertype, if so then we match
            location=None
            if servicetype is None or servicetype in rsp:
                print("Response matches server type: %s"%(servicetype))
                # use the message source address as default location
                location=addr 
                # extract the location or IP from the message
                for line in rsp.split("\r\n"): # cut into lines
                    tmp=line.split(":", 1) # cut line into key/val
                    # is this the key we care about -> then record the value
                    if len(tmp)>1 and tmp[0].lower()=="LOCATION".lower():
                        location=tmp[1].strip()
                        # strip http:// xxxx /stuff
                        if location.startswith("http://"):
                            location=location[7:] # strip http
                        if '/' in location:
                            location=location[:location.index('/')] # strip desc.xml
                        print("Got location: %s"%(location))
                        break # done with this response
                # add to the list of possible servers
                print("Loc added to response list: %s"%(location))
                responses.append(location)
        except socket.timeout:
            print("Socket timeout")
        if len(responses)>0: break
    return responses

class UtopiaClient:
    """Class for managing a client connection to a UtopiaServer."""

    UTOPIA_SSDP_SERVICE="utopia/1.1"
    DEFAULTHOST='localhost'
    DEFAULTPORT=8400    
    HEARTBEATINTERVAL_ms=1000
    HEARTBEATINTERVALUDP_ms=200
    MAXMESSAGESIZE = 1024 * 1024

    def __init__(self, clientstate=None):
        self.isConnected = False
        self.sock = []
        self.udpsock=None
        self.recvbuf = b''
        self.tsClock = TimeStampClock()
        self.sendHeartbeats = True
        self.nextHeartbeatTime = self.getTimeStamp()
        self.nextHeartbeatTimeUDP = self.getTimeStamp()
        self.ssdpDiscover = None
        self.clientstate = clientstate

    # time-stamp management
    def setTimeStampClock(self, tsClock):
        if not hasattr(tsClock,'getTimeStamp'):
            raise ValueError("Time Stamp clock must have getTimeStamp method")
        self.tsClock = tsClock
    def getTimeStamp(self):
        """Get the time-stamp for the current time"""
        return self.tsClock.getTimeStamp()
    def disableHeartbeats(self):
        '''' stop sending hearbeat messages. Use, e.g. when you want to use your own time-stamp clock. '''
        self.sendHeartbeats = False
    def enableHeartbeats(self):
        self.sendHeartbeats = True

    def connect(self, hostname=None, port=8400):
        """connect([hostname, port]) -- make a connection, default host:port is localhost:1972"""
        if hostname is None:       hostname = UtopiaClient.DEFAULTHOST
        if port is None or port<0: port     = UtopiaClient.DEFAULTPORT
        print("Trying to connect to: %s:%d"%(hostname, port))
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((hostname, port))
        self.sock.setblocking(False)
        self.sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1) # disable nagle
        self.isConnected = True
        # make udp socket also
        self.udpsock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0)
        # ensure tcp and udp have the same local port number...
        self.udpsock.bind(self.sock.getsockname())
        return self.isConnected

    def autoconnect(self, hostname=None, port=None, timeout_ms=5000, 
                queryifhostnotfound=False, localhostifhostnotfound=True, scanifhostnotfound=False):
        if port is None: port = UtopiaClient.DEFAULTPORT
        if hostname is None:
            print('Trying to auto-discover the utopia-hub server')
            if True:
                if self.ssdpDiscover is None:
                    print('making discovery object')
                    from mindaffectBCI.ssdpDiscover import ssdpDiscover
                    # create the discovery object
                    self.ssdpDiscover=ssdpDiscover(UtopiaClient.UTOPIA_SSDP_SERVICE)
                hosts=self.ssdpDiscover.discover(timeout=timeout_ms/1000.0)
            else:
                hosts=ssdpDiscover(servicetype=UtopiaClient.UTOPIA_SSDP_SERVICE, timeout=5, numretries=int(max(1, timeout_ms/5000)))
            print("Discovery returned %d utopia-hub servers"%len(hosts))
            if( len(hosts)>0 ):
                hostname=hosts[0].strip()
                print('Discovered utopia-hub on %s ...'%(hostname))
            if hostname is None:
                print('Error:: couldnt autodiscover the decoder!')


        if hostname is not None:
            self.try_connect(hostname,port,timeout_ms)

        # Try different ways of getting the host info: localhost, query, scan
        if not self.isConnected and localhostifhostnotfound : 
            self.try_connect('localhost',port,timeout_ms)
        
        if not self.isConnected and queryifhostnotfound:
            # ask user for host
            print("Could not auto-connect.  Trying manual")
            hostname = input("Enter the hostname/IP of the Utopia-HUB: ")
            self.try_connect(hostname,port,timeout_ms)

        if not self.isConnected and scanifhostnotfound:
            print("Could not auto-discover.  Trying IP scan.")
            from mindaffectBCI.ssdpDiscover import ipscanDiscover
            hosts = ipscanDiscover(port)
            if  len(hosts)>0:
                hostname=hosts[0].strip()
                print('Discovered utopia-hub on %s ...'%(hostname))
                self.try_connect(hostname,port,timeout_ms)
            
        if not self.isConnected:            
            raise socket.error('Connection Refused!')

    def try_connect(self, hostname, port=None, timeout_ms=5000):
        if ":" in hostname:
            hostname, port=hostname.split(":")
            port=int(port)

        for i in range(max(1, int(timeout_ms/1000))):
            try:
                print("Tring to connect to: %s:%d"%(hostname, port))
                self.connect(hostname, port)
                print('Connected!', flush=True)
                break
            except socket.error as ex:
                print('Connection refused...  Waiting', flush=True)
                print(ex)
                time.sleep(1)


    def gethostport(self):
        if self.isConnected:
            hp=self.sock.getpeername()
            return ":".join(str(i) for i in hp)
        return None
        
    def disconnect(self):
        """disconnect() -- close a connection."""
        if self.isConnected:
            self.sock.close()
            self.sock = []
            self.isConnected = False

    def sendRaw(self, request):
        """Send all bytes of the string 'request' out to socket."""
        if not(self.isConnected):
            raise IOError('Not connected to utopia server')

        N = len(request)
        nw = self.sock.send(request)
        while nw<N:
            nw += self.sock.send(request[nw:])

    def sendRawUDP(self, request):
        self.udpsock.sendto(request, self.sock.getsockname())

    def sendMessage(self, msg):
        if not msg is RawMessage: # convert to raw for sending
            if msg.timestamp is None: # insert valid time-stamp, N.B. all messages have timestamp!
                msg.timestamp = self.getTimeStamp()
            msg = RawMessage.fromUtopiaMessage(msg)
        self.sendRaw(msg.serialize()) # send the raw message directly
        self.sendHeartbeatIfTimeout()

    def sendHeartbeatIfTimeout(self, timestamp=None):
        if not self.sendHeartbeats:
            return
        if timestamp is None:
            timestamp = self.getTimeStamp()
        if timestamp > self.nextHeartbeatTime:
            self.nextHeartbeatTime=timestamp+self.HEARTBEATINTERVAL_ms
            self.sendRaw(RawMessage.fromUtopiaMessage(Heartbeat(timestamp,self.clientstate)).serialize())
        if timestamp > self.nextHeartbeatTimeUDP:
            self.nextHeartbeatTimeUDP=timestamp+self.HEARTBEATINTERVALUDP_ms
            self.sendRawUDP(RawMessage.fromUtopiaMessage(Heartbeat(timestamp,self.clientstate)).serialize())
        
    def sendMessages(self, msgs):
        """sends single or multiple utopia-messages to the utopia server
        """
        for msg in msgs:
            self.sendMessage(msg)
            
    def recvall(self, timeout_ms=0):
        """Read all the data from the socket immeaditely or block for timeout_ms if nothing to do."""
        if timeout_ms>0:
            self.sock.setblocking(1)
            self.sock.settimeout(timeout_ms/1000.0)
        else:
            self.sock.setblocking(0)
        data=[]
        try:
            data = self.sock.recv(self.MAXMESSAGESIZE)
        except socket.timeout:
            pass
        except socket.error as ex:
            if not ( ex.errno == 11 or ex.errno == 10035 ): # 11 is raised when no-data to read
                print("Socket error" + str(ex) + "#" + str(ex.errno) ) 
        return bytes(data)
    
    def getNewMessages(self, timeout_ms=250):
        """Wait for new messages from the utopia-server (with optional timeout), decode them and return the list of new messages"""
        # get all the data in the socket
        buf = self.recvall(timeout_ms)
        # append to the receive buffer
        self.recvbuf = self.recvbuf + buf
        # decode all the new messages to RawMessage
        (newmessages, nconsumed) = RawMessage.deserializeMany(self.recvbuf)
        self.recvbuf = self.recvbuf[nconsumed:] # remove the consumed part of the buffer
        # decode the RawMessages to actual messages
        newmessages = decodeRawMessages(newmessages)
        self.sendHeartbeatIfTimeout()
        return newmessages

    def initClockAlign(self, delays_ms=[50]*10):
        """Send some initial heartbeat messages to seed the alignment of the server clock with
        our local clock"""
        if not self.sendHeartbeats:
            print("Warning: not sending heartbeats as they are disabled!")
            return
        self.sendMessage(Heartbeat(self.getTimeStamp(),self.clientstate))
        for delay in delays_ms:
            time.sleep(delay/1000)
            self.sendMessage(Heartbeat(self.getTimeStamp(),self.clientstate))
    
    def messagelogger(self, timeout_ms=1000):
        """Simple message logger, infinite loop waiting for and printing messages from the server"""
        client.sendMessage(Subscribe(None, "ABCDEFGHIJKLMNOPQRSTUVWXYZ")) # subcribe to everything...
        while True:
            newmessages = client.getNewMessages(timeout_ms)
            print("%d new messages:"%len(newmessages))
            print("\n".join(str(msg) for msg in newmessages))


    def messagePingPong(self, timeout_ms=500):
        """Testing system sending 20 messages and printing any responses """
        # loop waiting for messages and printing them
        for i in range(20):
            self.sendMessage(StimulusEvent(i, [0, 1, 2, 3], [i+1, i+2, i+3, i+4]))
            time.sleep(1)
            newmessages = self.getNewMessages(timeout_ms)
            print("%d) %d NewMessages\n"%(i, len(newmessages)))
            print("\n".join(str(msg) for msg in newmessages))

            
def testSerialization():
    rm=RawMessage(1, 0, "payload")
    print("RawMessage: %s"%(rm))
    print("serialized: %s"%(rm.serialize()))
    print("deserialized: %s"%(RawMessage.deserialize(rm.serialize())[0]))
    print()

    hb=Heartbeat(10)
    print("Heartbeat: %s"%(hb))
    print("serialized: %s"%(hb.serialize()))
    print("deserialized: %s"%(Heartbeat.deserialize(hb.serialize())[0]))
    print()
    
    se=StimulusEvent(11, [0, 1, 2], [0, 1, 0])
    print("StimulusEvent: %s"%(se))
    print("serialized: %s"%(se.serialize()))
    print("deserialized: %s"%(StimulusEvent.deserialize(se.serialize())[0]))
    print()

    mc=ModeChange(12, "hello")
    print("ModeChange: %s"%(mc))
    print("serialized: %s"%(mc.serialize()))
    print("deserialized: %s"%(ModeChange.deserialize(mc.serialize())[0]))
    print()

    nt=NewTarget(13)
    print("NewTarget: %s"%(nt))
    print("serialized: %s"%(nt.serialize()))
    print("deserialized: %s"%(NewTarget.deserialize(nt.serialize())[0]))
    print()
  
    rs=Reset(14)
    print("Reset: %s"%(rs))
    print("serialized: %s"%(rs.serialize()))
    print("deserialized: %s"%(Reset.deserialize(rs.serialize())[0]))
    print()

    pt=PredictedTargetProb(15, 10, .5)
    print("PredictedTargetProb: %s"%(pt))
    print("serialized         : %s"%(pt.serialize()))
    print("deserialized: %s"%(PredictedTargetProb.deserialize(pt.serialize())[0]))
    print()

    pd=PredictedTargetDist(15, [1, 2, 3], [.5, .3, .2])
    print("PredictedTargetDist: %s"%(pd))
    print("serialized         : %s"%(pd.serialize()))
    print("deserialized: %s"%(PredictedTargetDist.deserialize(pd.serialize())[0]))
    print()
    
    sq=SignalQuality(16, [.1, .2, .3, .4, .5])
    print("SignalQuality: %s"%(sq))
    print("serialized   : %s"%(sq.serialize()))
    print("deserialized: %s"%(SignalQuality.deserialize(sq.serialize())[0]))
    print()
    
    dp=DataPacket(16, [[.1, .2, .3], [.4, .5, .6]])
    print("DataPacket : %s"%(dp))
    print("serialized : %s"%(dp.serialize()))
    print("deserialized: %s"%(DataPacket.deserialize(dp.serialize())[0]))
    print()
    

    rhb=RawMessage.fromUtopiaMessage(hb)
    print("Raw(Heartbeat): %s"%(rhb))

    srhb=rhb.serialize()
    print("serialized   : %s"%(srhb))
    dsrhb, =RawMessage.deserialize(srhb)
    print("Deserialized serialized Raw(Heartbeat): %s"%(dsrhb))
    sedsrhb=decodeRawMessage(dsrhb)
    print("Decoded deserialized Raw(Heartbeat): %s"%(sedsrhb))

    rse=RawMessage.fromUtopiaMessage(se)
    print("Raw(StimulusEvent): %s"%(rse))
    srse=rse.serialize()
    print("serialized       : %s"%(srse))    
    #deserialize
    (dsrse, nconsumed)=RawMessage.deserialize(srse)
    print("Deserialized serialized Raw(StimulusEvent): %s"%(dsrse))
    sedsrse=decodeRawMessage(dsrse)
    print("Decoded deserialized Raw(StimulusEvent): %s"%(sedsrse))

    #deserialize multiple
    (msgs, nconsumed)=RawMessage.deserializeMany(srse+srse+srhb)
    print("Deserialized Multiple: %s"%("\n".join([str(msg) for msg in msgs])))
    msgs=decodeRawMessages(msgs)
    print("Decoded Multiple: %s"%("\n".join([str(msg) for msg in msgs])))



def testSending():
    client = UtopiaClient()        
    client.autoconnect()

    # DataPacket
    client.sendMessage(DataPacket(client.getTimeStamp(), [[.1, .2, .3]]))

    # StimulusEvent
    time.sleep(5)

    
if __name__ == "__main__":

    #testCases()

    #hosts=ssdpDiscover(servicetype="ssdp:all", timeout=3)
    #print(hosts)

    # Just a small logging demo for testing purposes...
    hostname = None
    port = None
    
    if len(sys.argv)>1:
        hostname = sys.argv[1]
        tmp=[hostname.split(":")]
        if len(tmp)>1:
            hostname=tmp[0]
            port    =int(tmp[1])

    client = UtopiaClient()        
    client.autoconnect(hostname, port)
    client.messagelogger(30000)
    client.disconnect()



#!/bin/env python3

# Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from utopiaclient import *
from utopia2output import *

class Utopia2Cybathalon(Utopia2Output):    
    """Example class for a utopia OUTPUT module.  Connects to the utopia server
       and listens for output which exceed it's probability threshold before
       then printing them and using NEWTARGET to indicated the output has taken
       place. """
    
    def __init__(self,outputPressThreshold=.1,outputReleaseThreshold=None,myObjectIDs=None):
        super().__init__(outputPressThreshold,outputReleaseThreshold,myObjectIDs)

        self.br_player=1

        # Command offsets, do not change.
        self.CMD_SPEED= 1
        self.CMD_JUMP = 2
        self.CMD_ROLL = 3
        self.CMD_RST  = 99
        
        # Command configuration
        self.CMDS      = [self.CMD_ROLL, self.CMD_RST, self.CMD_JUMP, self.CMD_SPEED]
        self.THRESHOLDS= [.1,        .1,       .1,     .1      ]

        #Connect to BrainRacers
        self.br_hostname=None
        self.br_port=None

    def connect(self,host=UtopiaClient.DEFAULTHOST,port=UtopiaClient.DEFAULTPORT,br_host='localhost',br_port=5555,timeout_ms=30000):
        super().connect(host,port,timeout_ms)
        self.connecttocybathalon(br_host,br_port)
                    
    def doOutput(self,objID):
        """This function is run when objID has sufficiently low error to mean that 
           and output should be generated for this objID. 
           N.B. Override/Replace this function with your specific output method."""
        if self.myObjectIDs : 
            command = self.myObjectIDs.index(objID)
        else:
            command = objID

        if command : 
            print("Send cmd " + str(command) )
            self.sendtocybathalon(command)

    def connecttocybathalon(self,br_host,br_port):
        '''setup the connection to the cybathalon game'''
        self.br_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM);
        self.br_hostname=br_host
        self.br_port    =br_port
        print("Connecting to BrainRacers on: %s:%d"%(self.br_hostname,self.br_port))

            
    def sendtocybathalon(self,command):
        '''send the given command to the cybathalon game'''
        cmd = (self.br_player * 10) + command # command is [player,command] as number
        data = struct.pack('B', cmd)
        
        self.br_socket.sendto(data, (self.br_hostname, self.br_port))

            
''' simple driver for testing '''
if __name__ == "__main__":
    import sys
    host = UtopiaClient.DEFAULTHOST
    port = UtopiaClient.DEFAULTPORT
    br_host='localhost'
    br_port=5555

    if len(sys.argv)>1:
        hostname = sys.argv[1]
        tmp=hostname.split(":")
        if len(tmp)>1:
            hostname=tmp[0]
            port    =int(tmp[1])
    print('utopia: %s:%d'%(hostname,port))
            
    outputthreshold=.1
    if len(sys.argv)>2:
        try:
            outputthreshold = float(sys.argv[2])
        except:
            outputthreshold = None
            print('Non-numerical outputthreshold, using selection mode')
            
    if len(sys.argv)>3:
        br_host = sys.argv[3]
        print('br_host:br_port %s:%d'%(br_host,br_port))
        tmp=br_host.split(":")
        if len(tmp)>1:
            br_host =tmp[0]
            br_port =int(tmp[1])
    print('Cybathalon %s:%d'%(br_host,br_port))

    u2c = Utopia2Cybathalon(outputthreshold)
    u2c.connect(host,port,br_host,br_port)

    u2c.sendtocybathalon(u2c.CMD_SPEED)

    u2c.run()



import numpy as np
import warnings

def multipleCCA(Cxx=None, Cxy=None, Cyy=None,
                reg=0, regy=0,
                rank=1, CCA=True, rcond=1e-6, symetric=False):
    '''
    Compute multiple multi-way CCA (MCCA) decompositions using the given summary statistics
      [J,W,R]=multiCCA(Cxx,Cxy,Cyy,regx,regy,rank,CCA)
    Inputs:
      Cxx  = (d,d) current data cross-covariance (or any n-way symetric covariance matrix)
      Cxy  = (nM,nE,tau,d) current per output ERPs
      Cyy  = (nM,nE,tau,nE,tau) current response covariance for each output
      reg  = (1,) :float or (2,):float linear weighting reg strength or separate values for Cxx and Cyy 
            OR (d,) regularisation ridge coefficients for Cxx  (0)
      rank= [1x1] number of top cca components to return (1)
      CCA = [bool] or (2,):bool  flag if we normalize the X component, or the Y component. (true)
      rcond  = [float] tolerance for singular eigenvalues.        (1e-4)
               or (2,) separate rcond for Cxx (rcond(1)) and Cyy (rcond(2))
               or (2,1) (-1<-0) negative values = keep this fraction of eigen-values
               or (2,1) <-1 keep this many eigenvalues
      symetric = [bool] us symetric whitener?
    Outputs:
      J     = [nY x 1] optimisation objective scores
      W     = [d x rank x nY] spatial filters for each output
      R     = [tau x nE x rank x nY] responses for each stimulus event for each output
    Examples:
      # Supervised CCA
      Y = (nEp/nSamp,nY,nE) indicator for each event-type and output of it's type in each epoch
      X = (nEp/nSamp,tau,d) sliced pre-processed raw data into per-stimulus event responses
      stimTimes = (nEp) # sample number for each epoch or None if sample rate
      Cxx,Cxy,Cyy=updateSummaryStatistics(X, Y, stimTimes)
      J,w, r=multiCCA(Cxx, Cxy, Cyy, reg) # w=(nM,d)[dx1],r=(nM,nE,tau)[tau x nE] are the found spatial,spectral filters
      # incremental zeroTrain
      Cxx=[];Cxy=[];Cyy=[]; prediction=[];
      while isCalibration:
         # Slicing
         newX,newE,stimTimes= preprocessnSliceTrial(); #function which get 1 trials data+stim
         # Model Fitting
         Cxx, Cxy, Cyy= updateSummaryStatistics(newX, newE, stimTimes, Cxx, Cxy, Cyy);
         J, w, r      = multipleCCA(Cxx, Cxy, Cyy, regx, regy) # w=[dx1],r=[tau x nE] are the found spatial,spectral filters
      end
      # re-compute with current state, can try different regulisors, e.g. for adaptive bad-ch
      regx = updateRegFromBadCh();
      J, w, r=multipleCCA(Cxx, Cxy, Cyy, reg)
    Copyright (c) MindAffect B.V. 2018
    '''
    rank = int(max(1, rank))
    if not hasattr(reg, '__iter__'):
        reg = (reg, reg)  # ensure 2 element list
    if not hasattr(CCA, '__iter__'):
        CCA = (CCA, CCA)  # ensure 2 element list
    if not hasattr(rcond, '__iter__'):
        rcond = (rcond, rcond)  # ensure 2 element list
            
    # convert to double precision if needed
    if np.issubdtype(Cxx2d.dtype, np.float32) or np.issubdtype(Cxx2d.dtype, np.signedinteger):
        Cxx2d = np.array(Cxx2d, dtype=np.float64)
    if np.issubdtype(Cxy2d.dtype, np.float32) or np.issubdtype(Cxy2d.dtype, np.signedinteger):
        Cxy2d = np.array(Cxy2d, dtype=np.float64)
    if np.issubdtype(Cyy2d.dtype, np.float32) or np.issubdtype(Cyy2d.dtype, np.signedinteger):
        Cyy2d = np.array(Cyy2d, dtype=np.float64)
    
    # do CCA/PLS for each output in turn
    J = np.zeros((Cxy2d.shape[0]))  # [ nM ]
    W = np.zeros((Cxy2d.shape[0], rank, Cxy2d.shape[2]))  # (nM,rank,d)
    R = np.zeros((Cxy2d.shape[0], rank, Cxy2d.shape[1]))  # (nM,rank,(nE*tau)) 
    for mi in range(Cxy2d.shape[0]):  # loop over posible models
        # get output specific ERPs
        Cxym = Cxy2d[mi, :, :]  # ((tau*nE),d)

        # Whitener for X
        if CCA[0]:
            # compute model specific whitener, or re-use the  last one
            if mi <= Cxx2d.shape[0]:
                isqrtCxx, _ = robust_whitener(Cxx2d[mi, :, :], reg[0], rcond[0], symetric)
            # compute whitened Cxy
            isqrtCxxCxym = np.dot(Cxym, isqrtCxx) # (nM,(nE*tau),d)
        else:
            isqrtCxxCxym = Cxym

        # Whitener for Y
        if CCA[1]:
            # compute model-specific whitener, or re-use the last one
            if mi <= Cyy2d.shape[0]:
                isqrtCyy, _ = robust_whitener(Cyy2d[mi, :, :], reg[1], rcond[1], symetric)
            isqrtCxxCxymisqrtCyy = np.dot(isqrtCyy.T, isqrtCxxCxym)
        else:
            isqrtCxxCxymisqrtCyy = isqrtCxxCxym

        # SVD for the double  whitened cross covariance
        Rm, lm, Wm = np.linalg.svd(isqrtCxxCxymisqrtCyy, full_matrices=False)  #N.B. Rm=((nE*tau),rank),lm=(rank),Wm=(rank,d)
        Wm = Wm.T  # (d,rank)

        # include relative component weighting directly in the  Left/Right singular values
        nlm = lm / np.max(lm)  # normalize so predictions have unit average norm
        Wm = Wm * np.sqrt(nlm[np.newaxis, :])
        Rm = Rm * np.sqrt(nlm[np.newaxis, :])

        # pre-apply the pre-whitener so can apply the result directly on input data
        if CCA[0]:
            Wm = np.dot(isqrtCxx, Wm)
        if CCA[1]:
            Rm = np.dot(isqrtCyy, Rm)

        # extract the desired parts of the solution information, largest singular values first
        slmidx = np.argsort(lm)  # N.B. ASCENDING order
        slmidx = slmidx[::-1]  # N.B. DESCENDING order
        r = min(len(lm), rank)  # guard rank>effective dim
        W[mi, :r, :] = Wm[:, slmidx[:r]].T  #(nM,rank,d)
        J[mi] = lm[slmidx[-1]]  # N.B. this is *wrong* for rank>1
        R[mi, :r, :] = Rm[:, slmidx[:r]].T  #(nM,rank,(nE*tau))
        
    # Map back to input shape
    if Cyy.ndim > 2:
        R = np.reshape(R, (R.shape[0], R.shape[1], Cyy.shape[-2], Cyy.shape[-1])) # (nM,rank,nE,tau)
    if Cxx.ndim > 2:
        W = np.reshape(W, (W.shape[0], W.shape[1], Cxx.shape[-2], Cxx.shape[-1]))

    # strip model dim if not needed
    if Cxy.ndim == 3:
        R = R[0, ...]
        W = W[0, ...]
        J = J[0, ...]
        
    return J, W, R


def robust_whitener(C:np.ndarray, reg:float=0, rcond:float=1e-6, symetric:bool=True, verb:int=0):
    """compute a robust whitener for the input covariance matrix C, s.t. isqrtC*C*isqrtC.T = I
    Args:
        C ((d,d) np.ndarray): Sample covariance matrix of the data
        reg (float, optional): regularization strength when computing the whitener. Defaults to 0.
        rcond (float, optional): reverse-condition number for truncating degenerate eigen-values when computing the inverse. Defaults to 1e-6.
        symetric (bool, optional): flag to produce a symetric-whitener (which preserves location) or not. Defaults to True.
        verb (int, optional): verbosity level. Defaults to 0.

    Returns:
        W (np.ndarray): The whitening matrix
        iW (np.ndarray): The inverse whitening matrix
    """    

    # ensure symetric
    C = (C + C.T) / 2

    # include the regularisor if needed
    # TODO[]: include the ridge later, i.e. after the eigendecomp?
    if not reg is None and not reg == 0:
        # TODO[]: Optimal shrinkage?
        if np.ndim(reg) == 0 or len(reg) == 1:  # weight w.r.t. same amplitude identity
            C = (1-reg)*C + reg*np.eye(C.shape[0], dtype=C.dtype)*np.mean(C.diagonal())
        elif len(reg) == C.shape[0]:  # diag entries
            np.fill_diagonal(C, C.diagonal()+reg)
        else:  # full reg matrix
            C = C + reg

    # eigen decomp
    sigma, U = np.linalg.eig(C)  # sigma=(r,) U=(d,k)
    U = U.real
    
    # identify bad/degenerate eigen-values, complex, negative, inf, nan or too small
    bad = np.any(np.vstack((np.abs(sigma.imag > np.finfo(sigma.dtype).eps),  # complex
                            np.isinf(sigma),  # inf
                            np.isnan(sigma),  # nan
                            np.abs(sigma)<np.finfo(sigma.dtype).eps)), 0)  # too-small
    if verb:
        print("{} bad eig".format(sum(bad)))
    # zero these bad ones
    sigma[bad] = 0

    # additional badness conditions based on eigen-spectrum
    if not rcond is None:
        # identify eigen-values we want to remove due to rcond
        if 0 <= rcond:  # value threshold
            bad = np.logical_or(bad, sigma.real < rcond*max(np.abs(sigma)))
        elif -1 < rcond and rcond < 0:  # fraction
            si = np.argsort(sigma)  # N.B. Ascending
            bad[si[:int(len(si)*abs(rcond))]] = True  # up to this fraction are bad
        elif rcond < -1:  # number to discard
            si = np.argsort(sigma)
            bad[si[:abs(rcond)]] = True  # this many are bad

    if verb:
        print("{} rcond+bad eig".format(sum(bad)))

    # compute the whitener (and it's inverse)
    if not all(bad):
        keep = np.logical_not(bad)
        Ukeep = U[:, keep]  # (d,r)
        sqrtsigmakeep = np.sqrt(np.abs(sigma[keep]))  #(r,)
        # non-symetric (and rank  reducing) version
        sqrtC = Ukeep * sqrtsigmakeep[np.newaxis, :]
        isqrtC= Ukeep * (1.0/sqrtsigmakeep[np.newaxis, :])
        
        # post apply U to get symetric version if wanted
        if symetric:
            sqrtC = np.dot(sqrtC, Ukeep.T)
            isqrtC= np.dot(isqrtC, Ukeep.T)
    else:
        warnings.warn('Degenerate C matrices input!')
        sqrtC = 1
        isqrtC = 1
    return (isqrtC, isqrtC)


def plot_multicca_solution(w, r):
    # plot the solutions, spatial-filter / temporal-impulse-response
    import matplotlib.pyplot as plt
    plt.clf()
    plt.subplot(211); plt.plot(np.squeeze(w)); plt.title("Spatial")
    plt.subplot(212); plt.plot(np.squeeze(r).T); plt.title("Temporal")
    plt.show()


def cvSupervised(Xe, Me, stimTimes, evtlabs=('re', 'fe'), n_splits=10, rank=1):
    ''' do a cross-validated training of a multicca model using sklearn'''
    from mindaffectBCI.decoder.updateSummaryStatistics import updateSummaryStatistics, crossautocov, updateCyy, updateCxy, autocov, updateCxx
    from mindaffectBCI.decoder.multipleCCA import multipleCCA
    from mindaffectBCI.decoder.scoreOutput import scoreOutput, dedupY0
    from mindaffectBCI.decoder.scoreStimulus import scoreStimulus, scoreStimulusCont
    from mindaffectBCI.decoder.decodingCurveSupervised import decodingCurveSupervised
    from mindaffectBCI.decoder.decodingSupervised import decodingSupervised
    from mindaffectBCI.decoder.stim2event import stim2event
    from sklearn.model_selection import StratifiedKFold
    print("Xe = {}\nMe = {}".format(Xe.shape, Me.shape))
    # convert from stimulus coding to brain response coding
    Ye = stim2event(Me, evtlabs, axis=1) # (nTrl, nEp, nY, nE) [ nE x y x ep x trl ]

    # cross validate in folds over trials
    Ye_true = Ye[:, :, 0:1, :] # N.B. horrible slicing trick to keep the dim
    print("Ye = {}\nYe_true = {}".format(Ye.shape, Ye_true.shape))
    Fy = np.zeros((1, Ye.shape[0], Ye.shape[1], Ye.shape[2])) # (nModel, nTrl, nEp, nY) [ nY x nEp x nTrl x nModel ]
    print("Fy = {}".format(Fy.shape))

    kf = StratifiedKFold(n_splits=n_splits)
    train_idx = np.arange(30)
    valid_idx = np.arange(30, Ye.shape[0])
    print("CV:", end='')
    for i, (train_idx, valid_idx) in enumerate(kf.split(np.zeros(Xe.shape[0]), np.zeros(Ye_true.shape[0]))):
        print(".", end='', flush=True)
        Cxx, Cxy, Cyy = updateSummaryStatistics(Xe[train_idx, :, :, :], Ye_true[train_idx, :, :, :], stimTimes)
        J, W, R = multipleCCA(Cxx, Cxy, Cyy, rank=rank)
        # valid performance
        Fei = scoreStimulus(Xe[valid_idx, :, :, :], W, R)
        #print("Fei = {}".format(Fei.shape))
        Fyi = scoreOutput(Fei, Ye[valid_idx, :, :, :], dedup0=True)
        #print("Fyi = {}".format(Fyi.shape))
        Fy[:, valid_idx, :, :] = Fyi
    decodingCurveSupervised(Fy)

    # retrain with all the data
    Cxx, Cxy, Cyy = updateSummaryStatistics(Xe, Ye_true, stimTimes)
    J, W, R = multipleCCA(Cxx, Cxy, Cyy)
    return (W, R, Fy)
    

def testcase():
    from mindffectBCI.decoder.utils import testNoSignal, testSignal, sliceData, sliceY
    #from multipleCCA import *
    if False:
        X, Y, st = testNoSignal()
    else:
        X, Y, st, A, B = testSignal(tau=10, noise2signal=10)
    # TODO[]: summary stats directly without first slicing
    Y_true = Y[:, :, 0:1, :] if Y.ndim > 3 else Y[:, 0:1, :] # N.B. hack with [0] to stop removal of dim...
    from mindaffectBCI.decoder.updateSummaryStatistics import updateSummaryStatistics, plot_summary_statistics
    Cxx, Cxy, Cyy = updateSummaryStatistics(X, Y_true, tau=10)

    #plot_summary_statistics(Cxx,Cxy,Cyy)

    # single supervised training
    from mindaffectBCI.decoder.multipleCCA import multipleCCA, plot_multicca_solution
    J, w, r = multipleCCA(Cxx, Cxy, Cyy, rcond=-.3, symetric=True) # 50% rank reduction
    
    plot_multicca_solution(w, r)

    # apply to the data
    from scoreStimulus import scoreStimulus
    Fe = scoreStimulus(X, w, r)
    import matplotlib.pyplot as plt;
    plt.clf();
    plt.plot(np.einsum("Tsd,d->s",X,w.ravel()),'b',label="Xw");
    plt.plot(Y_true.ravel(),'g',label="Y");
    plt.plot(Fe.ravel()/np.max(Fe.ravel())/2,'r',label="Fe");
    plt.legend()
    from scoreOutput import scoreOutput
    print("Fe={}".format(Fe.shape))
    print("Y={}".format(Y.shape))
    Fy = scoreOutput(Fe, Y, r) # (nM,nTrl,nEp,nY)
    print("Fy={}".format(Fy.shape))
    import matplotlib.pyplot as plt
    sFy=np.cumsum(Fy, -2)
    plt.clf();plt.plot(sFy[0, 0, :, :]);plt.xlabel('epoch');plt.ylabel('output');plt.show()

    from decodingSupervised import decodingSupervised
    decodingSupervised(Fy)
    from decodingCurveSupervised import decodingCurveSupervised
    decodingCurveSupervised(Fy)

if __name__=="__main__":
    testcase()
